                ;=================================;
                ;                                 ;
                ;       WATERING CONTROLLER       ;
                ;                                 ;
                ;=================================;

; For all information on this program, please contact:
;
; Jean-Philippe CHEVREAU
; 22, chemin de la Pinede
; 06650 Le Rouret
; FRANCE
;
; tel - home: (33) 493 094 677


; Last modified: 03 DEC 96

;---------------------------------
; Display messages are in English
;---------------------------------

;*****************************************************************************
;
; Features:
; ---------
;
; This watering controller controls 8 watering zones, where each zone
; consists of a number of sprinklers controlled by usually one valve,
; sometimes more (the maximum number of valves in a zone  depends on the
; hardware used to control the valves). In other words, this controller
; has 8 valve control outputs or channels.
;
; This watering controller features 2 watering programs per day. In each
; program:
;
;  - the duration of each zone can be set from 0 to 99 minutes
;    in 1 minute increments;
;
;  - up to two start times can be defined in hours and minutes.
;
; On top of this, a weekly scheduling system permits to enable/disable
; each of the two programs individually for each day of the week.
;
;
; Example: (in this example, Zones #4, #7 and #8 are not used)
;
; Program 1 = start at 05h30 and 23h15, activate Zone #1 for 20 minutes,
;             Zone #2 for 15 minutes and Zone #5 for 12 minutes;
; Program 2 = start at 20h20, activate Zone #1 for 25 minutes,
;             Zone #3 for 10 minutes and Zone #6 for 12 minutes;
; Weekly Scheduling:
;   Monday and Thursday: run Program 1 and Program 2
;   Tuesday and Friday: run Program 1
;   Wednesday and Saturday: run Program 2
;   Sunday: no watering
;
; Note: since we have two programs with two start times each, we can have
;       up to four start times in one day. However, the weekly scheduling
;       does not permit to enable/disable each the two start times of a
;       program individually.
;
;-----------------------------------------------------------------------------
;
; During an outage, the Watering Controller will not operate, but a
;  backup battery and a power down feature will enable the Watering
;  Controller to be operational again as soon as the outage is over.
;
; The display is an 8 character alphanumeric display, so the user interface
;  is friendly. This display is not an LCD but is based on LEDs, so it is
;  readable even in the dark. The power down feature automatically blanks
;  the display during an outage, so the draw on the backup battery will be
;  very low, which permits to save user information even if an outage
;  lasts several hours or if outages are frequent.
;
; If the backup battery is flat, then in case of outage, the Watering
;  Controller will perform a reset when the power comes back. All user
;  information will be lost, but a Default Program will ensure a minimum
;  amount of watering. This is useful for instance if you go away to
;  vacation for a long time. If the backup battery is flat, the garden
;  will not stay for many days without any watering at all. Naturally,
;  since the current time is also lost, this minimum watering service
;  may occur at any time of the day or of the night. The Default Program
;  can be disabled by changing a strap on the printed circuit (or with
;  a switch, depending on the hardware implementation).
;
; An input permits to take into account the information given by a
;  humidity probe. If this input is low, watering programs are disabled and
;  a humidity message ("HUMID") is displayed in alternance with the Clock.
;  Note: only the state of the probe at the start time of a program is
;  taken into account. Once a program has started, it ignores the state
;  of the probe.
;
; A test function permits to activate each zone individually, which
;  makes it easy to adjust the sprinklers of each zone without having
;  to make a special watering program. Just in case, the duration for each
;  zone is limited to 60 minutes (but normally the test will be stopped
;  manually). The test function ignores the state of the humidity probe.
;
;*****************************************************************************
;
; Naming conventions
; ------------------
;
; ZONE = Zone (refers to the valve and associated group of sprinklers)
; ZONE1 = Zone #1, etc.
;
; In what follows, the word "program" ususally refers to a watering program,
;  rather than the microcontroller program. A watering program consists of
;  two start times (hours + minutes) and a duration (in minutes) for each of
;  the 8 zones.
;
; PR1 = Program 1. Starts at (PR1A_HOUR, PR1A_MINU) and (PR1B_HOUR, PR1B_MINU)
;                  Controls zones ZONE1 to ZONE8 for durations of PR1_ZONE1
;                  to PR1_ZONE8 minutes respectively.
; PR2 = Program 2. Starts at (PR2A_HOUR, PR2A_MINU) and (PR2B_HOUR, PR2B_MINU)
;                  Controls zones ZONE1 to ZONE8 for durations of PR2_ZONE1
;                  to PR2_ZONE8 minutes respectively.
;
;*****************************************************************************
;
; User's Guide
; ------------
;
; There are four push-buttons:
;
;      /-----\      /-----\      /-----\      /-----\
;      | PB1 |      | PB2 |      | PB3 |      | PB4 |
;      \-----/      \-----/      \-----/      \-----/
;
;       Menu        Submenu        Set         Stop
;
; 1.- Depressing PB1 will take us to the next menu if no program is
;     currently active. If a program is active, it must be aborted using
;     the "Manual Stop" command (see item 6 below). The menus are:
;       0:  Menu Idle (= not in a menu)
;       1:  Set Clock
;       2:  Set PR1 Start Times
;       3:  Set PR1 Durations
;       4:  Set PR2 Start Times
;       5:  Set PR2 Durations
;       6:  Set Weekly Schedule
;       7:  Test Sprinklers
;
; Notes:
;
;     - Depressing PB1 (going to the next menu) always resets the submenu.
;
;     - A Watering Program may run only when not in a menu ("Menu Idle").
;
; 2.- Depressing PB2 will take us to the next submenu. The submenus depend on
;     which menu we are currently in. When we are not in a menu (Menu Idle),
;     PB2 alone has no effect.
;
;     The current Menu/Submenu combination is made visible by the fact that
;     the number which can be set is flashing. This number will not change
;     as long as PB3 is not depressed.
;
;     In the "Set Clock" menu, the submenus are:
;       0:  Set Days
;       1:  Set Hours
;       2:  Set Minutes
;
;     In the "Set PR1/PR2 Start Times" menus, the submenus are:
;       0:  Set Hours of Start Time A
;       1:  Set Minutes of Start Time A
;       2:  Set Hours of Start Time B
;       3:  Set Minutes of Start Time B
;
;     In the "Set PR1/PR2 Durations" menus, the submenus are:
;       0:  Set Duration of ZONE1
;       1:  Set Duration of ZONE2
;       2:  Set Duration of ZONE3
;       3:  Set Duration of ZONE4
;       4:  Set Duration of ZONE5
;       5:  Set Duration of ZONE6
;       6:  Set Duration of ZONE7
;       7:  Set Duration of ZONE8
;
;     In the "Set Weekly Schedule" menu, the submenus are:
;       0:  Set Monday Schedule
;       1:  Set Tuesday Schedule
;       2:  Set Wednesday Schedule
;       3:  Set Thursday Schedule
;       4:  Set Friday Schedule
;       5:  Set Saturday Schedule
;       6:  Set Sunday Schedule
;
;           Note: each schedule has four possible values, corresponding
;           to the following commands:
;               NONE_SCH = 0: disable PR1 and PR2
;               PR1_SCH  = 1: enable PR1, disable PR2
;               PR2_SCH  = 2: enable PR2, disable PR1
;               BOTH_SCH = 3: enable PR1 and PR2
;
;     In the "Test Sprinklers" menu, the submenus are:
;       0:  Test ZONE1
;       1:  Test ZONE2
;       2:  Test ZONE3
;       3:  Test ZONE4
;       4:  Test ZONE5
;       5:  Test ZONE6
;       6:  Test ZONE7
;       7:  Test ZONE8
;
; 3.- When in a Menu, depressing PB3 will increment the number selected by
;     the menu and the submenu at a rate of 1 unit every 400 ms as long as
;     PB3 stays depressed. This permits to set the number to the desired
;     value.
;
;     Depressing PB3 briefly increments the number selected once without
;     having to keep PB3 depressed for 400 ms. This makes it easy to reach
;     the exact desired value.
;
; 4.- Depressing PB2 as well as PB3 will decrement the number selected
;     by the menu and the submenu instead of incrementing it. This is handy
;     when the maximum value which can be set is high since it takes a long
;     time to go round all the values (in particular for the durations which
;     can go from 0 to 99 minutes).
;
; 5.- Depressing PB4 as well as PB3 will set the number selected by the
;     menu and the submenu to zero (Monday if the value is the day of the
;     week). This is handy in particular to clear the duration of a zone
;     that is no longer used.
;
; 6.- Depressing PB4 will take us to Menu Idle (i.e. will take us out of
;     the menus) immediately no matter where we are in the menus/submenus
;     (and so will de-activate any zone under test).
;
; 7.- Operation of the Test Sprinklers menu: this menu is a bit different
;     from the others because it permits immediate watering whereas the
;     other menus only permit to set up the condition under which a
;     program will operate.
;
;     The Test Sprinklers menu is identified by the "TS" keyword on the first
;     two digits of the Display. To test a given zone, you must first select
;     the zone by depressing PB2, and then start watering by depressing PB3.
;     The fact that watering is not started as soon as the zone is selected
;     permits to select, say, Zone #7 without starting watering (even briefly)
;     in any of the previous zones. 
;
;     During watering the Display will indicate the zone number and the
;     remaining time like in program execution. The time per zone is fixed to
;     60 minutes. This is just a safeguard, because the test is meant to be
;     stopped manually, either by depressing PB4 to exit the Test Sprinklers
;     menu completely, either by depressing PB2 to test the next zone.
;
; 8.- "Manual Start" commands:
;
;     Depressing PB1 and PB2 simultaneously when we are in Menu Idle
;     will abort any program that may be active and will start program PR1.
;
;     Similarly, depressing PB1 and PB3 simultaneously when we are in
;     Menu Idle will abort any program that may be active and will start
;     program PR2.
;
;     These "Manual Start" commands permit to start a program immediately
;     without changing the start times, nor the schedule, i.e. without
;     reprogramming the Watering Controller. Starting a program manually
;     does not inhibit the automatic start for that day, so these commands
;     must be viewed as a means to adding more watering respective to what
;     was programmed. For instance, if the start times of PR1 are later
;     than then end of PR1 started manually, PR1 will also be started
;     automatically. Note that if PR1 is already active when it is started
;     manually, it will be restarted from the beginning.
;
;     These commands are ignored when we are in a menu.
;
; 9.- "Manual Stop" command: depressing PB4 when we are in Menu Idle
;     will abort the currently active Watering Program, if any. It does not
;     matter whether the Program was started automatically or manually.
;
;10.- "Schedule All". This command is provided to speed up the scheduling,
;     in particular for users who want the same watering every day.
;
;     Depressing PB1 and PB2 simultaneously when we are in the "Set Weekly
;      Schedule" menu will change the schedule for all days of the week at
;      the same time. Thus, depressing PB1 and PB2 several times will set
;      the following commands for all days of the week:
;       - 1st time: enable PR1, disable PR2 (schedule value = PR1_SCH)
;       - 2nd time: enable PR2, disable PR1 (schedule value = PR2_SCH)
;       - 3rd time: enable PR1 and PR2      (schedule value = BOTH_SCH)
;       - 4th time: disable PR1 and PR2     (schedule value = NONE_SCH)
;
;     Naturally, any previous scheduling with differences depending on the
;      day of the week will be lost.
;
;11.- "Skip Zone" command. Depressing PB2 when a program is running will
;     make the program go to the next zone immediately or to terminate if
;     it was the last zone. Note that the action of PB2 in this situation
;     is similar to the case of the "Test Sprinklers" menu.
;
;-----------------------------------------------------------------------------
;
; State a power up:
; -----------------
;
; In all cases, the clock is set to Monday 12h00 (midday) and the start
;  times of the two programs are reset to 0h00. Additionally:
;
; 1) if the DEFIN input (P0.2) is low, the default program is disabled:
;    all the durations are reset to zero and the two programs are
;    descheduled for all days of the week;
;
; 2) if the DEFIN input (P0.2) is high, the default program is enabled:
;    the durations of the eigth zones of Program #1 are set to 10 minutes,
;    and Program #1 is scheduled for all days of the week. Program #2 is
;    mot used (its durations are reset to zero and it is descheduled for
;    all days of the week).
;
; Note: the reason why the clock is set to 12h00 and the start times of the
;       default program (PR1) are reset to 0h00 is to prevent the Watering
;       Controller from watering as soon as it is powered up when the
;       default program is enabled.
;
; The first time power is applied, or if there is an outage and the backup
;  battery is flat, the 8 characters of the Display will flash to signal
;  this condition. This flashing will stop as soon as any push-button is
;  depressed.
;
; The 6th character of the Display (the ':' of the Clock) will flash to
;  indicate that an outage occurred and that the backup battery operated
;  correctly. This flashing will stop as soon as any push-button is
;  depressed.
;
;*****************************************************************************
;
; Display Patterns
; ----------------
;
; Note: In a menu, the value which can be set (depending on the
;  submenu) flashes.
;
;
; 1.- In Menu Idle and in the Set Clock menu
;
;   ---------------------------------
;   | M | O |   | 1 | 2 | : | 0 | 0 |
;   ---------------------------------
;     Days        Hours      Minutes
;
;   Note: the clock display is on 24 hours (8 PM is displayed as 20:00).
;
;
; 2.- In the Set PR1 / PR2 Start Times menus
;
;   ---------------------------------
;   | P | 1 | A | 0 | 6 | : | 3 | 0 |
;   ---------------------------------
;    Program  ^      Start Time
;    Number   |  (Hours)     (Minutes)
;             |
;             \ (A or B)
;
;
; 3.- In the Set PR1 / PR2 Durations menus
;
;   ---------------------------------
;   | P | 1 |   | Z | 1 |   | 1 | 0 |
;   ---------------------------------
;    Program       Zone      Duration
;    Number       Number     (Minutes)
;
;
; 4.- In the Set Weekly Schedule menu
;
;   ---------------------------------
;   | M | O |   | P | 1 |   | - | - |
;   ---------------------------------
;     Days      Program #1  Program #2
;
;             Enabled: P1    Enabled: P2
;             Disabled: --   Disabled: --
;
;
; 5.- In the Test Sprinklers menu, before starting watering
;
;   ---------------------------------
;   | T | S |   | Z | 1 |   | 0 | 0 |
;   ---------------------------------
;     Test         Zone 
;   Indicator     Number
;
;
; 6.- In the Test Sprinklers menu, after starting watering
;
;   ---------------------------------
;   | T | S |   | Z | 1 |   | 1 | 0 |
;   ---------------------------------
;     Test         Zone      Remaining
;   Indicator     Number     Duration
;                            (Minutes)
;
;
; 7.- When a Program is running
;
;   ---------------------------------
;   | P | 1 |   | Z | 1 |   | 1 | 0 |
;   ---------------------------------
;    Program       Zone      Remaining
;    Number       Number     Duration
;                            (Minutes)
;
;
; 8.- Humidity message
;
;   ---------------------------------
;   | H | U | M | I | D |   |   |   |
;   ---------------------------------
;
;*****************************************************************************
;
; Effect of push-buttons depending on current state of Watering Controller
; ------------------------------------------------------------------------
;
; Since there are many push-button combinations and they can have different
;  effects depending on the current state of the Watering Controller, it is
;  useful to recap them here.
;
; Notes:
;
;  1) Push-button combinations which are not quoted are simply ignored
;     by the Watering Controller.
;
;  2) "Controller Idle" means: neither executing a Program, nor in a Menu.
;
;
;  Button         | Controller Idle   | In a Menu          | Program Active
;  combination    |                   |                    |
; ----------------+-------------------+--------------------+------------------
;                 |                   |                    |
;  PB1 (Menu)     | Enter Menus       | Go to next Menu    | No effect
;                 |                   |                    |
;  PB2 (Submenu)  | No effect         | Go to next Submenu | Skip Zone
;                 |                   |                    |
;  PB3 (Set)      | No effect         | Increment value    | No effect
;                 |                   |                    |
;  PB4 (Stop)     | No effect         | Exit Menus         | Stop Program
;                 |                   |                    |
;  PB1 & PB2      | Start PR1         | In Set Weekly      | Stop Program
;                 | (manual start)    |  Schedule menu,    |  and start PR1
;                 |                   |  schedule all;     |
;                 |                   | else no effect     |
;                 |                   |                    |
;  PB1 & PB3      | Start PR2         | No effect          | Stop Program
;                 | (manual start)    |                    |  and start PR2
;                 |                   |                    |
;  PB2 & PB3      | No effect         | Decrement value    | No effect
;                 |                   |                    |
;  PB3 & PB4      | No effect         | Clear value        | No effect
;
;
;*****************************************************************************
;
; Program overlapping
; -------------------
;
; When we are executing one of the programs, PR1 or PR2, DURA_ACTIVE
;  is non-zero and we stop looking for the start times of any programs.
;  So, if the duration of a program Px is such that it is not finished at
;  the time the other program Py should start, then Py will never start.
;  For instance (assuming only Start Time A is used), if the start time
;  of PR1 is 04h00 and each of the 8 zones is programmed for 15 minutes,
;  then the start time of PR2 should be 06h00 at the earliest. Similarly,
;  if the distance between Start Times A and B of a program is smaller
;  than the total duration of that program, then Start Time B will be
;  lost (assuming B is later than A).
;
; In particular, it is possible to disable the 'B' Start Times by making
;  them identical to the corresponding 'A' Start Times. This fact is
;  turned to good account in the way we make the 'B' Start Times
;  transparent to users who don't need them (see paragraph "Choice
;  of a solution to make 'B' Start Times transparent" below).
;
; Note that it is OK for two programs to be "adjacent", that is, the
;  minimum amount of time between the end of a program and the start time
;  of the next program (or Start Time B respective to Start Time A and
;  vice-versa) is zero.
;
;*****************************************************************************
;
; Choice of a solution to make 'B' Start Times transparent
; --------------------------------------------------------
;
; Presumably many users will use only one start time per program.
;  To disable the 'B' Start Times, they must be made identical to the
;  corresponding 'A' Start Times. If the two groups of start times (A and B)
;  were totally independent, the user of the 'A' Start Times only would have
;  to modify the 'B' Start Times every time he modifies the 'A' Start Times.
;  Three solutions can be considered to solve this problem:
;
; 1 - Every time Start Time A is modified, the corresponding Start Time B is
;     made identical to Start Time A.
;
; 2 - Every time Start Time A is modified, the corresponding Start Time B is
;     made identical to Start Time A as long as Start Time B was the same
;     as Start Time A initially. Once Start Time B is made different, the
;     changes on Start Time A have no influence on it.
;
; 3 - Every change in Start Time A is duplicated in the corresponding Start
;     Time B, so that the interval between the two remains constant.
;
; First note that the three solutions are identical as long as you never
; change a 'B' Start Time. The drawback of the first solution is that every
; time Start Time A is modified, Start Time B is smashed if it was different.
; This solution is fine for those who don't use the 'B' Start Times, but is
; a pain for the others. The second solution is a good compromise between
; users and non-users of the 'B' Start Times. The third solution is
; interesting because it permits to move the two start times (for a given
; program) without changing the interval between the two. Whether this is
; a better solution for the user than having two independent groups once
; they have been decoupled as in the second solution, is debatable. Anyway,
; we chose the third solution.
;
; Notes about the solution chosen (#3):
;
; 1) Although the 'Set Start Time B' menu can be viewed as setting the
;    interval between the two groups (whereas the 'Set Start Time A' menu
;    sets the position of the two groups as a whole), what is displayed in
;    the 'Set Start Time B' menu (in hours and minutes) is the absolute start
;    time, not the interval.
;
; 2) Although the 'B' Start Time is modified by the 'Set Start Time A' menu,
;    only the 'A' Start Time is displayed in this menu. To view the
;    corresponding change in the 'B' Start Time, you must enter the 'Set Start
;    Time B' menu (even though you may no need to change Start Time B).
;
; 3) When Start Time B is "pushed" by Start Time A, this is done modulo 24
;    hours. For instance, if Start Time A = 12 h and Start Time B = 22 h
;    initially, and if Start Time A is set to 16 h (+ 4 hours), then Start
;    Time B will be set to 2 h without any change in the Weekly Schedule.
;    So if the program in question was enabled on Monday and disabled on
;    Tuesday, nothing will be done to enable it on Tuesday although Start
;    Time B was moved from 22 h to 2 h. This would get too confusing, and
;    in fact we can't even do it as long as the scheduling is on a per
;    program basis, i.e. you can't enable Start Time A and disable Start
;    Time B of a given program and vice versa.
;
; 4) When setting Start Time A, the interval between start times A and B is
;    kept constant as long as increment (PB3) and decrement (PB2 and PB3)
;    commands are used. If a clear command is used (PB3 and PB4), the value
;    (either hours or minutes) will be set to zero in both Start Time A and
;    Start Time B.
;
;*****************************************************************************
;
; Notes on the control of the 8-Character Display
; -----------------------------------------------
;
; We use an HDSP-2111 Eight Character Smart 5*7 Alphanumeric Display
;  (Hewlett-Packard). The display colour is yellow. Naturally, it makes
;  no difference to the program to use another colour (HDSP-2110: orange,
;  HDSP-2112: red, HDSP-2113: green).
;
; As explained in the paragraph about Port Usage below, we shall output
;  the data in series and convert it back to parallel with an external
;  8-bit shift register (74HC164) to save I/O pins. The update of the Display
;  will be somewhat slower, but this does not matter for our application.
;
; Since we'll never need to read the Display, we can tie RD/ to VDD and tie
;  together WR/ and CE/.
;
; We tie CLS to VDD to use the internal clock. We do not use the clock output
;  CLK.
;
; We tie A4 to VDD because we shall not use the UDC feature (User Defined
;  Characters). So we use A0 to A3 (we need A3 because we need to write into
;  the Control Word Register).
;
; See the initialization section (Label Ini_Dis) for comment on the value
;  written into the Control Word Register.
;
; There must be a delay of at least 110 us after the reset before accessing
;  the Display. The time taken to clear the RAM after the reset and before
;  entering the background loop is sufficient, since it is 382 us.
;  (With a 6 MHz CPU clock, clearing one RAM location takes 2+4 = 6 us,
;  so clearing 63 locations takes 378 us, plus 4 us to initialize the loop).
;
;*****************************************************************************
;
; Port Usage
; ----------
;
; P0.0 - HUMP, Humidity Probe input - low to disable watering programs
;
; P0.1 = MPIN, Mains Power input - used to blank the Display upon loss
;        of mains power to save backup battery
;        * low = mains power off
;        * high = mains power on
;
; P0.2 = DEFIN, Enable Default Program input
;
;
; To save I/O pins, we output the Valve control signals in multiplexed form,
;  i.e. one "data" bit, VOUT/ (the multiplexed Valve control output) and
;  3 address bits (ZA0 to ZA2). An external de-multiplexer (74HC138)
;  delivers a control signal to each Valve control circuit. In any case,
;  only one Valve control output (corresponding to a zone) can be active
;  at any time.
;
; P1.0 = ZA0, Zone address output, bit 0
; P1.1 = ZA1, Zone address output, bit 1
; P1.2 = ZA2, Zone address output, bit 2
; P1.3 = VOUT/, Multiplexed Valve control output - active low
;
; N.B. The multiplexed Valve control output is active low so that
;      it is inactive upon a reset.
;
; P1.4 = PB1/, Menu push-button     - active low
; P1.5 = PB2/, Sub-menu push-button - active low
; P1.6 = PB3/, Set push-button      - active low
; P1.7 = PB4/, Stop push-button     - active low
;
;
; Since there are not enough I/O pins left to drive all the signals of the
;  Display (8 data pins + 4 address pins + flash input + write strobe,
;  i.e. 14 pins), we shall output the data in series and convert it back
;  to parallel with an external 8-bit shift register (74HC164). This saves
;  6 I/O pins since we replace the 8 data pins with one serial output and
;  one serial clock and it enables us to control the whole Display with
;  only one 8-bit port:
;
; P3.0 = DA0  = A0 of HDSP-2111
; P3.1 = DA1  = A1 of HDSP-2111
; P3.2 = DA2  = A2 of HDSP-2111
; P3.3 = DA3  = A3 of HDSP-2111
; P3.4 = DFL/ = FL/ of HDSP-2111
; P3.5 = DWR/ = WR/ & CE/ of HDSP-2111 tied together
; P3.6 = DSC  = Serial Clock
; P3.7 = DSD  = Serial Data
;
;*****************************************************************************
;
; RAM Map
; -------
;
; 1 - Register Banks
;
;     We only use Register Bank 0 (we never change the Register Bank,
;     so we always use the default which is 0). Therefore registers
;     R0 to R7 are always equivalent to addresses 00h to 07h.
;
; 2 - Stack Area
;
;     The Stack Pointer is initialized to 07h by default.
;     We decide to reserve the 08h - 11h area (10 bytes) for the stack.
;     This is sufficient as explained below, so we'll be able
;     to place some variables in the 12h - 1Fh area.
;
;     Stack usage:
;
;     The largest subroutine nesting level is 4 (see the Subroutine Nesting
;     Tree at the end of the file): the Background calls Ctrl_Dur (1st level)
;     which in turn calls Dis_Dur (2nd level) which in turn calls Dis_Pr
;     (3rd level) which in turn calls Out_Char (4th level).
;     This takes 4 * 2 = 8 bytes on the Stack.
;
;     Additionally, the Timer 0 interrupt subroutine will take 2 bytes
;     to save the PC (no registers are saved). Thus the maximum stack usage
;     is 8 + 2 = 10 bytes.
;
; 3 - Bit Addressable Area
;
;     Let alone the flags (packed in one byte at address 20h), we need
;     bit access in the following variables:
;       - FLASH_STA
;       - PROG_ACTIVE
;       - SECONDS
;
;     This is why these variables have been placed in the bit addressable
;     area (20h - 2Fh). Note that the other variables that we have placed
;     in this area do not require bit access.
;
;*****************************************************************************
;
; Timing
; ------
;
; All timings (main clock, watering durations, incrementation of a parameter
;  selected in a menu, push-button debouncing) are based on a 10-millisecond
;  tick given by Timer 0.
;
; Clock fine tuning
; -----------------
;
; This feature must be modified (or removed) depending on the crystal being
;  used. The 3 crystals (of the same origin) that we tried all gave a slighty
;  too high frequency, making the Clock lead by about 5.3 s per day, which
;  corresponds to a frequency too high by 368 Hz (+ 61 ppm). This may also
;  be due to the capacitors associated with the crystal circuitry (2 x 33 pF).
;
; Anyway, we did not make a hardware fix but a firmware fix by skipping
;  a 10-millisecond tick every CLOCK_ADJ seconds, where CLOCK_ADJ is
;  parameter whose maximum value is 255. This permits compensation for
;  crystal having a frequency too high by at least 39 ppm (1/25,500).
;
; If the particular crystal that you have does not meet the condition that
;  the frequency is too high by at least 39 ppm (corresponding to a lead of
;  3.4 s per day), then it is not sufficient to change the value of CLOCK_ADJ.
;  Unfortunately, there is no RAM left to count a longer period than 255
;  seconds to remove a tick, so I suggest to change the Timer 0 counter to
;  count 4999 cycles instead of 5000 (see "Timer 0 Configuration" below).
;  Naturally if the crystal frequency is too low instead of too high,
;  the solution is to add one extra tick every CLOCK_ADJ seconds.
;
;*****************************************************************************
;
; CPU = 87C750
; CPU Clock = 6.000 MHz
; Machine cycle = 2 us
;
;*****************************************************************************


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


$MOD750
$DEBUG

;*****************************************************************************

; Equates
; -------

;--------------- Clock fine tuning -------------------------------------------

        ; Without compensation, the Clock leads by about LEAD seconds per day.
        ;  Since we skip one tick every CLOCK_ADJ seconds, in one day we
        ;  skip 86,400 / CLOCK_ADJ ticks, i.e. 864 / CLOCK_ADJ seconds.
        ;  To perform the compensation, the number of seconds skipped must
        ;  be equal to LEAD. Hence:
        ;
        ;    864                               864
        ;  --------- = LEAD   =>   CLOCK_ADJ = ----
        ;  CLOCK_ADJ                           LEAD
        ;
        ; A lead of 5.3 s per day yields CLOCK_ADJ = 163.
        ; The best value in our case seems to be CLOCK_ADJ = 164
        ; (corresponding to a lead of 5.27 s per day).
        ;
        ; Note: the minimum lead which can be compensated for in this way is
        ;       864 / 255 = 3.4 s (corresponding to +39 ppm).

CLOCK_ADJ       EQU     164         ;Compensates for a lead of 5.27 s per day

;--------------- Other timing parameters -------------------------------------

DBNC_CNT        EQU     4           ;Number of times through the debouncing
                                    ; loop
DBNC_DLY        EQU     3           ;Elementary debouncing delay in units of
                                    ; 10 ms, hence 30 ms
; Note: the total debouncing delay is between DBNC_CNT * DBNC_DLY and
;  (DBNC_CNT + 1) * DBNC_DLY (i.e. between 120 and 150 ms).

INC_DLY         EQU     40          ;Delay between two incrementations (or
                                    ; decrementations) in units of 10 ms,
                                    ; hence 400 ms

;--------------- Push-Buttons definitions ------------------------------------

BUTTON_MSK      EQU     0F0h        ;Mask for the push-buttons (bits 4-7)

                ; Push-buttons positions
NOPB            EQU     0F0h        ;No button depressed
MENU            EQU     0E0h        ;PB1 (Menu) is depressed
                                    ; Used to: - go to next menu
SUBMENU         EQU     0D0h        ;PB2 (Submenu) is depressed
                                    ; Used to: - go to next submenu
                                    ;          - skip zone when program running
SETVAL          EQU     0B0h        ;PB3 (Set) is depressed
                                    ; Used to: - increment value to set
STOP            EQU     70h         ;PB4 (Stop) is depressed
                                    ; Used to: - exit Menus
                                    ;          - manually stop a Program
MST_PR1         EQU     0C0h        ;PB1 and PB2 are depressed
                                    ; Used to: - manually start PR1
                                    ;          - schedule all
MST_PR2         EQU     0A0h        ;PB1 and PB3 are depressed
                                    ; Used to: - manually start PR2
DOWNVAL         EQU     90h         ;PB3 and PB2 are depressed
                                    ; Used to: - decrement value to set
CLRVAL          EQU     30h         ;PB3 and PB4 are depressed
                                    ; Used to: - clear value to set

;--------------- Miscellaneous definitions -----------------------------------

N_ZONE          EQU     8           ;Number of Zones

MENU_MOD        EQU     8           ;Modulo value for MENU_STA
                                    ; = number of menus (0 to 7)

PR1_DIS         EQU     '1'         ;Value to display "P1" (for PR1)
PR2_DIS         EQU     '2'         ;Value to display "P2" (for PR2)
TST_DIS         EQU     'S'         ;Value to display "TS" (for Test)

PR1_SCH_MSK     EQU     01h         ;Mask to test PR1 schedule
PR2_SCH_MSK     EQU     02h         ;Mask to test PR2 schedule

PR1_ENA         EQU     ACC.0       ;Bit 0 of schedule (SCH_MON, etc.) is set
                                    ; when PR1 is enabled for that day
PR2_ENA         EQU     ACC.1       ;Bit 1 of schedule (SCH_MON, etc.) is set
                                    ; when PR2 is enabled for that day

NONE_SCH        EQU     00h         ;Value of schedule to disable PR1 and PR2
PR1_SCH         EQU     01h         ;Value of schedule to enable PR1 only
PR2_SCH         EQU     02h         ;Value of schedule to enable PR2 only
BOTH_SCH        EQU     03h         ;Value of schedule to enable PR1 and PR2

SCH_MOD         EQU     4           ;Modulo value for schedule

;--------------- HDSP-2111 8-Character Display definitions -------------------

CWR_ADDR        EQU     10h         ;Addr of Control Word Register of Display

                ; Addresses of Flashing status of characters of the Display
DIS_FL1         EQU     08h         ;Address of Flashing status of 1st char
DIS_FL2         EQU     09h         ;Address of Flashing status of 2nd char
DIS_FL3         EQU     0Ah         ;Address of Flashing status of 3rd char
DIS_FL4         EQU     0Bh         ;Address of Flashing status of 4th char
DIS_FL5         EQU     0Ch         ;Address of Flashing status of 5th char
DIS_FL6         EQU     0Dh         ;Address of Flashing status of 6th char
DIS_FL7         EQU     0Eh         ;Address of Flashing status of 7th char
DIS_FL8         EQU     0Fh         ;Address of Flashing status of 8th char

                ; Addresses of ASCII codes of characters of the Display
DIS_CH1         EQU     18h         ;Address of 1st character of Display
DIS_CH2         EQU     19h         ;Address of 2nd character of Display
DIS_CH3         EQU     1Ah         ;Address of 3rd character of Display
DIS_CH4         EQU     1Bh         ;Address of 4th character of Display
DIS_CH5         EQU     1Ch         ;Address of 5th character of Display
DIS_CH6         EQU     1Dh         ;Address of 6th character of Display
DIS_CH7         EQU     1Eh         ;Address of 7th character of Display
DIS_CH8         EQU     1Fh         ;Address of 8th character of Display

                ; Values to flash various groups of characters
                ; (using Flash function of HDSP-2111: bit 3 of CWR)
FLASH_1_2       EQU     03h         ;Flash 1st and 2nd characters
FLASH_4_5       EQU     18h         ;Flash 4th and 5th characters
FLASH_7_8       EQU     0C0h        ;Flash 7th and 8th characters
FLASH_4_8       EQU     0D8h        ;Flash 4th, 5th, 7th & 8th chars
NO_FLASH_B6     EQU     20h         ;Stop flashing all characters but the 6th

;--------------- I/O pins definitions ----------------------------------------

HUMP            EQU     P0.0        ;Humidity Probe Input
MPIN            EQU     P0.1        ;Mains Power Input
DEFIN           EQU     P0.2        ;Enable Default Program Input

VOUT            EQU     P1.3        ;Multiplexed Valve Control Output

DWR             EQU     P3.5        ;WR/ & CE/ of Display tied together
DSC             EQU     P3.6        ;Serial Clock of Display
DSD             EQU     P3.7        ;Serial Data of Display

;--------------- Register definitions ----------------------------------------

        ;If the following register definitions are changed, make sure
        ; to change the corresponding "_B" definitions as well
        ; (BUTTON_POS_B, etc.)

; R0 is a general purpose register, used as:
            ; - Indirect access to clear RAM upon reset
            ; - Base address of PR1 or PR2 start times when setting start time
            ;    (M2 and M4)
            ; - Address of character in the Display, input parameter of
            ;    Out_Char, Dis_Pr and BCD_ASC
            ; - Address of data to set, increment or decrement, input
            ;    parameter of Set_Val, Inc_BCD and Dec_BCD
            ; - Address of schedule in Sch_All and Dis_Sch
            ; - Address of Start Time when checking start times (Pr_Stt)
            ; - Address of duration of the zone currently active (NProg, DProg)
            ; - Address of duration to print duration in Dis_Dur
; R1 is a general purpose register, used as:
            ; - Loop counter in Push-buttons debouncing
            ; - Address of Duration in Set Durations and Test Sprinklers Menus,
            ;    and when a program is active (input parameter of Dis_Dur)
            ; - Address of Schedule, input parameter of Dis_Sch
            ; - Address of Hours of Start Time, input parameter of Dis_Stt
            ; - Address of Schedule when checking start times (SStt)

BUTTON_POS      EQU     R2          ;Position of push buttons
MENU_STA        EQU     R3          ;Menu State
SUBMENU_STA     EQU     R4          ;Submenu State
PR_DIS          EQU     R5          ;Program number or Test message
                                    ; (for display)
; R6 is a general purpose register, used as:
            ; - Saved position of push buttons for debouncing
            ; - Message index to display humidity message
            ; - Day of Week, input parameter of Dis_Day
            ; - BCD number, input parameter of BCD_ASC
            ; - Mask to check Schedule when checking start times (Pr_Stt, SStt)
            ; - Remaining duration when a program is active
            ;    or during sprinkler test
; R7 is a general purpose register, used as:
            ; - Character data, input parameter of Out_Char
            ; - Modulo value, input parameter of Set_Val, Inc_BCD and Dec_BCD
            ; - Total duration, input parameter of Ctrl_Dur

;-----------------------------------------------------------------------------

; Variable Definitions
; --------------------

; The following definitions are used when the instruction demands a
;  direct address rather than the register number (works because we
;  only use Register Bank 0)

R_0             DATA    00h         ;General purpose
R_1             DATA    01h         ;General purpose
BUTTON_POS_B    DATA    02h         ;Equivalent to BUTTON_POS  (i.e. R2)
MENU_STA_B      DATA    03h         ;Equivalent to MENU_STA    (i.e. R3)
SUBMENU_STA_B   DATA    04h         ;Equivalent to SUBMENU_STA (i.e. R4)
PR_DIS_B        DATA    05h         ;Equivalent to PR_DIS      (i.e. R5)
R_6             DATA    06h         ;General purpose
R_7             DATA    07h         ;General purpose

; Add the following definitions for debugging (easier than the full names)

R_2             DATA    02h         ;Corresponds to BUTTON_POS
R_3             DATA    03h         ;Corresponds to MENU_STA
R_4             DATA    04h         ;Corresponds to SUBMENU_STA
R_5             DATA    05h         ;Corresponds to PR_DIS


STACK           DATA    08h         ;Defined for debugging

DIS_CWR         DATA    12h         ;Image of Control Word Register of Display
OUT_BITS        DATA    13h         ;Loop counter in Out_Char subroutine

SUBMENU_MOD     DATA    14h         ;Modulo value for SUBMENU_STA (depends
                                    ; on the current value of MENU_STA)
ZONE_NUM        DATA    15h         ;Zone number (during program execution and
                                    ; during sprinkler test)
                                    ;Also used in Set Weekly Schedule menu to
                                    ; hold the scheduling value of the
                                    ; Schedule All command.
MINS            DATA    16h         ;Minutes, for zone duration
SEC_ADJ         DATA    17h         ;Seconds, for Clock fine tuning (not BCD)

; The start times (PR1A_HOUR to PR2B_MINU) must be contiguous in memory

PR1A_HOUR       DATA    18h         ;Start time A of PR1 - Hours
PR1A_MINU       DATA    19h         ;Start time A of PR1 - Minutes
PR1B_HOUR       DATA    1Ah         ;Start time B of PR1 - Hours
PR1B_MINU       DATA    1Bh         ;Start time B of PR1 - Minutes

PR2A_HOUR       DATA    1Ch         ;Start time A of PR2 - Hours
PR2A_MINU       DATA    1Dh         ;Start time A of PR2 - Minutes
PR2B_HOUR       DATA    1Eh         ;Start time B of PR2 - Hours
PR2B_MINU       DATA    1Fh         ;Start time B of PR2 - Minutes

; Flags (use addresse 20h) - - - - - - - - - - - - - - - - - - - - - - - - - -

FLAGS           DATA    20h         ;Reserved for flags

SET_ACTIVE      EQU     FLAGS.0     ;Set when user depresses Set button (PB3)
SET_DOWN        EQU     FLAGS.1     ;Set when user depresses PB2 as well as PB3
FIRST_SET       EQU     FLAGS.2     ;Set when user depresses Set button, clred
                                    ; after first incrementation/decrementation
                                    ;Also used in the Test Sprinklers menu to
                                    ; control watering (watering is started by
                                    ; depressing Set button)
DLY_ACTIVE      EQU     FLAGS.3     ;Set when 400 ms delay is active
SCH_ALL_RDY     EQU     FLAGS.4     ;Set when schedule all value is incremented
INIT_ZONE       EQU     FLAGS.5     ;Set when zone is initialized
END_ZONE        EQU     FLAGS.6     ;Set when zone duration is over
GOT_TICK        EQU     FLAGS.7     ;Set when Timer 0 interrupt occurred
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

FLASH_STA       DATA    21h         ;Flashing status of characters
                            ; Bit 0 = 1 -> 1st character flashes
                            ; Bit 1 = 1 -> 2nd character flashes
                            ; ...
                            ; Bit 7 = 1 -> 8th character flashes

DURA_ACTIVE     DATA    22h         ;Points to the duration of the zone
                                    ; currently active
                            ; 0          = no program active
                            ; #PR1_ZONE1 = Zone #1 of PR1 active
                            ; #PR1_ZONE2 = Zone #2 of PR1 active
                            ; ...
                            ; #PR2_ZONE8 = Zone #8 of PR2 active

PROG_ACTIVE     EQU     DURA_ACTIVE.5 ;Set when a program is active
            ; This works because when no program is active, DURA_ACTIVE = 0
            ;  so its bit 5 = 0, and when a program is active, DURA_ACTIVE
            ;  has a value in the range 29h to 38h (address of PR1_ZONE1 to
            ;  address of PR2_ZONE8), so its bit 5 = 1. This trick saves code
            ;  space because to test if a program is active, we have to test
            ;  one bit (JB or JNB) instead of comparing a byte to zero (MOV
            ;  to Acc followed by JZ or JNZ).

HUNDTHS         DATA    23h         ;Hundredths of seconds, for debouncing
HUNDREDTHS      DATA    24h         ;Hundredths of seconds of Clock
SECONDS         DATA    25h         ;Seconds (Clock) - bit addressable
MINUTES         DATA    26h         ;Minutes (Clock)
HOURS           DATA    27h         ;Hours (Clock)
DAYS            DATA    28h         ;Days (Clock)

; The durations for each Program (PR1_ZONE1 to PR1_ZONE8 on the one hand,
;  and PR2_ZONE1 to PR2_ZONE8 on the other hand) must be contiguous in memory

PR1_ZONE1       DATA    29h         ;Duration of ZONE1 in PR1
PR1_ZONE2       DATA    2Ah         ;Duration of ZONE2 in PR1
PR1_ZONE3       DATA    2Bh         ;Duration of ZONE3 in PR1
PR1_ZONE4       DATA    2Ch         ;Duration of ZONE4 in PR1
PR1_ZONE5       DATA    2Dh         ;Duration of ZONE5 in PR1
PR1_ZONE6       DATA    2Eh         ;Duration of ZONE6 in PR1
PR1_ZONE7       DATA    2Fh         ;Duration of ZONE7 in PR1
PR1_ZONE8       DATA    30h         ;Duration of ZONE8 in PR1

PR2_ZONE1       DATA    31h         ;Duration of ZONE1 in PR2
PR2_ZONE2       DATA    32h         ;Duration of ZONE2 in PR2
PR2_ZONE3       DATA    33h         ;Duration of ZONE3 in PR2
PR2_ZONE4       DATA    34h         ;Duration of ZONE4 in PR2
PR2_ZONE5       DATA    35h         ;Duration of ZONE5 in PR2
PR2_ZONE6       DATA    36h         ;Duration of ZONE6 in PR2
PR2_ZONE7       DATA    37h         ;Duration of ZONE7 in PR2
PR2_ZONE8       DATA    38h         ;Duration of ZONE8 in PR2

; The schedules (SCH_MON to SCH_SUN) must be contiguous in memory

SCH_MON         DATA    39h         ;Monday Schedule
SCH_TUE         DATA    3Ah         ;Tuesday Schedule
SCH_WED         DATA    3Bh         ;Wednesday Schedule
SCH_THU         DATA    3Ch         ;Thursday Schedule
SCH_FRI         DATA    3Dh         ;Friday Schedule
SCH_SAT         DATA    3Eh         ;Saturday Schedule
SCH_SUN         DATA    3Fh         ;Sunday Schedule

; Warning: all the RAM is used!

;-----------------------------------------------------------------------------
; Note on the scope of Variables
;
; The following variables and registers must be preserved over a round
;  through the background:
;
; BUTTON_POS (R2), MENU_STA (R3), SUBMENU_STA (R4), PR_DIS (R5),
; DIS_CWR, SUBMENU_MOD, ZONE_NUM, HUNDTHS,
; PR1A_HOUR, PR1A_MINU, PR1B_HOUR, PR1B_MINU,
; PR2A_HOUR, PR2A_MINU, PR2B_HOUR, PR2B_MINU,
; SET_ACTIVE, SET_DOWN, FIRST_SET, DLY_ACTIVE, SCH_ALL_RDY,
; INIT_ZONE, END_ZONE, GOT_TICK,
; FLASH_STA, DURA_ACTIVE,
; MINS, HUNDREDTHS, SECONDS, MINUTES, DAYS, HOURS,
; PR1_ZONE1, PR1_ZONE2, PR1_ZONE3, PR1_ZONE4,
; PR1_ZONE5, PR1_ZONE6, PR1_ZONE7, PR1_ZONE8,
; PR2_ZONE1, PR2_ZONE2, PR2_ZONE3, PR2_ZONE4,
; PR2_ZONE5, PR2_ZONE6, PR2_ZONE7, PR2_ZONE8,
; SCH_MON, SCH_TUE, SCH_WED, SCH_THU, SCH_FRI, SCH_SAT, SCH_SUN.
;
; As a matter of fact, the only variables or registers which do not have
;  to be preserved over a round through the background are R0, R1, R6, R7
;  and OUT_BITS.
;
; Note: the "Side effects" section in subroutines and procedures only mentions
;       the Acc, B and the general purpose registers R0, R1, R6 and R7.
;-----------------------------------------------------------------------------


;*****************************************************************************
;
; Initializations on Reset
; ========================
;
;
; Special Function Registers
; --------------------------
;
; TCON register = 10h
; - - - - - - - - - -
;
; bit
;  7    GATE =  0   Timer/counter is enabled when TR = 1
;  6    C/T =   0   Timer operation from internal clock
;  5    TF =    0   (Will be set on overflow of Timer/Counter 0)
;  4    TR =    1   Timer/Counter 0 enabled
;  3    IE0 =   0   (INT0 is not used)
;  2    IT0 =   0   (INT0 is not used)
;  1    IE1 =   0   (INT1 is not used)
;  0    IT1 =   0   (INT1 is not used)
;
;
; IE register = 82h
; - - - - - - - - -
;
; bit
;  7    EA =    1   Enable Interrupts
;  6    -       0
;  5    -       0
;  4    -       0
;  3    -       0
;  2    EX1 =   0   Disable External Interrupt 1 (INT1)
;  1    ET0 =   1   Enable Timer 0 Overflow Interrupt
;  0    EX0 =   0   Disable External Interrupt 0 (INT0)
;
;
; Since a machine cycle = 2 us, to obtain a Timer 0 interrupt every 10 ms,
;  the counter must count 5000 = 1388h machine cycles. Since the counter
;  will count up from the reload value until it overflows (FFFFh -> 0000h),
;  the reload value must be 10000h - 1388h = EC78h.
;
;
; Initialization of Port3
; -----------------------
;
; It is necessary to initialize Port3 because it controls the 8-Character
;  Display:
;
;   - P3.5 = DWR/ must be set to high;
;   - P3.6 = DSC must be set to low so that we have a low to high transition
;            on the serial clock the first time we write data to the Display;
;   - the other bits are irrelevant at this point.
;
; Since the port outputs are initialized to 1's after a reset, all we have
;  to do is to clear DSC (P3.6).
;
;
; Initialization of 8-Character Display
; -------------------------------------
;
; At power up, the 8 characters will flash to indicate that a reset has
;  occurred. To do this, we use the "Blink" function of the Display
;  (bit 4 of Control Word Register). As soon as any push-button is
;  depressed (indicating a user intervention), the flashing is stopped.
;
; So we write the following value into the Control Word Register:
;
;  - Brightness is set to D2-D0 = 010 (53 % brightness);
;  - Flash function of HDSP-2111 is on, so D3 = 1;
;  - Blink function of HDSP-2111 (flash 8 characters) is on at power up,
;    so D4 = 1
;    (later when a push-button is depressed, Blink is off, so D4 = 0);
;  - Bit 5 is read only - assume D5 = 0;
;  - Self Test is not used, so D6 = 0;
;  - Clear Function is not used, so D7 = 0.
;
;  D7 D6 D5 D4 D3 D2 D1 D0
;   0  0  0  1  1  0  1  0  =  1Ah      (at power up)
;
;*****************************************************************************


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


;*****************************************************************************
;
; Reset Vector
; ------------
;
;*****************************************************************************

        ORG     0                   ;Reset vector
Reset:
            ; Initialization of SFRs
        MOV     TCON, #10h          ;Set up timer controls
        MOV     IE, #82h            ;Enable Timer 0 interrupt
        MOV     RTL, #78h           ;Low byte of Timer 0 reload value
            ; No more space in the reset vector area, instruction
            ;  to set RTH is placed at label Reset1
        AJMP    Reset1



;*****************************************************************************
;
; Timer 0 Interrupt Routine
; -------------------------
;
; This routine simply sets the GOT_TICK flag every 10 ms. The GOT_TICK flag
;  is tested and cleared in the Clock subroutine (called by the Background).
;
;*****************************************************************************

        ORG     0Bh                 ;Timer 0 interrupt
        SETB    GOT_TICK            ;Tell Clock subroutine tick occurred
        RETI



;*****************************************************************************
;
; Initializations on Reset continued
; ----------------------------------
;
;*****************************************************************************

Reset1:     ; Moved the instructions to set TCON, IE and RTL to the reset
            ;  vector area to save code space
        MOV     RTH, #0ECh          ;High byte of Timer 0 reload value

            ; Clear User RAM (00h - 3Fh).
            ; Use R0 = 00h to indirectly access RAM location.
            ; All the RAM except R0 itself (01h - 3Fh) is cleared using R0
            ;  as an index, but at the end of the process, R0 is also equal
            ;  to zero, so in fact all the RAM is cleared.
            ; This gives enough delay (382 us) for the 8-Character Display
            ;  which needs at least 110 us before being accessed after a reset.
        MOV     R0, #3Fh            ;Start with top of RAM
        CLR     A                   ;Acc used to write 0 in each RAM location
Clr_Ram:
        MOV     @R0, A              ;Clear one RAM location
        DJNZ    R0, Clr_Ram         ;Loop until RAM address = 00h

            ; Port3 initialization
        CLR     DSC                 ;Set Display Serial Clock to low state

Ini_Dis:    ; 8-Character Display initialization
        MOV     DIS_CWR, #1Ah       ;Write Control Word Register of Display
                                    ; (will be output to Display in Out_CWR)

            ; Specific initializations
        MOV     BUTTON_POS, #NOPB   ;Assume no button depressed to start with
        MOV     HOURS, #12h         ;Clock starts with Monday 12h00

            ; Default Watering Program
Mon_Def:
        JNB     DEFIN, ERes         ;If DEFIN low, Default Program is disabled
                        ; Duration of 8 zones of PR1 = 10 minutes
        MOV     R0, #PR1_ZONE1      ;Get start address of PR1 Durations area
        MOV     A, #10h             ;Duration for all 8 zones = 10 minutes
LDef:
        MOV     @R0, A              ;Write Duration for a zone
        INC     R0                  ;Go to next zone
        CJNE    R0, #PR1_ZONE8+1, LDef ;Loop until the 8 zones are done
                        ; Schedule PR1 for all days of the week
        MOV     A, #PR1_SCH         ;Value to enable PR1 (PR2 stays disabled)
        ACALL   Sch_All             ;Write value in the 7 schedules

ERes:
            ; Initializations over, fall through to background loop


;=============================================================================
;
;   Start of Background Loop
;
;=============================================================================

Bkgnd:

;*****************************************************************************
;
; Push-buttons debouncing
; -----------------------
;
; Whenever we detect a change in the position of the push-buttons
;  (saved in BUTTON_POS) we validate this change only if we read the
;  same value 4 more times at 30 ms intervals. Therefore the debouncing
;  delay is 120 to 150 ms.
;
; Note that we chose this relatively high debouncing delay to detect
;  unambiguously the fact that 2 buttons are depressed simultaneously.
;
; When any stable push-button combination is detected (which includes
;  depressing any of the 4 push-buttons individually), we clear both the
;  power-up indicator (flashing of the 8 characters) and the outage indicator
;  (flashing of the hours / minutes separator, i.e. the 6th character of the
;  Display).
;
; "Step-by-step" setting:
; - - - - - - - - - - - -
;
;   The number selected is incremented (resp. decremented) immediately when
;   PB2 (Set push-button) is depressed (resp. PB2 & PB3 together). The 400 ms
;   delay occurs only for the second incrementation (resp. decrementation).
;
;   In particular, this feature allows easy "step-by-step" setting of a
;   value, since depressing the Set push-button briefly increments the number
;   selected once without having to keep that button depressed for 400 ms.
;
;   This is done by setting flag FIRST_SET when the Set push-button
;   is depressed and by clearing it when the value is incremented once.
;   While FIRST_SET is set, the 400 ms delay is bypassed.
;
; "Skip Zone" command:
; - - - - - - - - - -
;
;   Depressing PB2 (Submenu push-button) when a program is running will
;   make the program go to the next zone immediately or to terminate if
;   it was the last zone.
;
;   This is done by clearing flag INIT_ZONE when PB2 is depressed. This
;   flag is tested in the "Execute Watering Program" section at label BProg.
;   When it is FALSE, we go the next zone without waiting for the end of
;   the duration.
;
; Special case of Test Sprinklers menu:
; - - - - - - - - - - - - - - - - - - -
;
;   When entering the Test Sprinklers menu (MENU_STA incremented from 6 to 7),
;   and when PB2 is depressed in the Test Sprinklers menu, we set flags
;   INIT_ZONE and END_ZONE because we don't want to start watering until
;   PB3 is depressed (see Notes in Test Sprinklers section). This trick makes
;   the Ctrl_Dur routine displays a duration of 00 as long as PB3 is not
;   depressed, since we are in the same state as when the duration is over.
;
;*****************************************************************************

        ;----------------------
        ; Debounce the buttons
        ;----------------------

        MOV     A, P1               ;Read current position of push-buttons
        ANL     A, #BUTTON_MSK      ;Mask bits other than push-buttons
        CJNE    A, BUTTON_POS_B, Dbnc ;If position has changed, debounce it
        AJMP    EButt

Dbnc:   MOV     R6, A               ;Save current position of push-buttons
        MOV     R1, #DBNC_CNT       ;Go four times through debouncing loop

LDbnc:  MOV     HUNDTHS, #0         ;Initialize debouncing delay

WDbnc:  ACALL   Clock               ;Execute Clock since this is a blind loop
        MOV     A, HUNDTHS          ;Wait 30 ms before reading
        CJNE    A, #DBNC_DLY, WDbnc ; push-buttons again

        MOV     A, P1               ;Read again current pos of push-buttons
        ANL     A, #BUTTON_MSK      ;Mask bits other than push-buttons
        CJNE    A, R_6, EButt       ;If position not stable, ignore change

        DJNZ    R1, LDbnc           ;Loop back if debouncing not complete

            ; New position is stable (5 times the same value at 30 ms
            ;  intervals), so it becomes the new validated position
        MOV     BUTTON_POS_B, R6

        ;---------------------------------------
        ; Interpret the new position of buttons
        ;---------------------------------------

                ; Depressing any push-button(s) will stop the flashing
                ;  due to a general reset or to an outage
        ANL     DIS_CWR, #0EFh      ;Stop flashing the 8 chars of the Display:
                                    ; clear bit 4 of Control Word Register
                                    ; (will be output to Display in Out_CWR)
        CLR     FLASH_STA.5         ;Stop flashing 6th character of Display

        CJNE    BUTTON_POS, #NOPB, SVal ;No button depressed: get ready
        CLR     SCH_ALL_RDY             ; to increment "schedule all" value

                ; Set button (PB3) is depressed
SVal:   CJNE    BUTTON_POS, #SETVAL, SDwn   ;PB3 is depressed
        CLR     SET_DOWN                    ;To increment value
        SJMP    SAct                        ;Go tell Set_Val it can increment

                ; PB2 is depressed as well as PB3
SDwn:   CJNE    BUTTON_POS, #DOWNVAL, SClr  ;PB2 and PB3 are depressed
        SETB    SET_DOWN                    ;To decrement value
        SJMP    SAct                        ;Go tell Set_Val it can decrement

                ; PB4 is depressed as well as PB3
SClr:   CJNE    BUTTON_POS, #CLRVAL, Men    ;PB3 and PB4 are depressed
                        ;Note: no flag is set here, BUTTON_POS will be compared
                        ; to #CLRVAL again in Set_Val

SAct:   SETB    SET_ACTIVE                  ;Tell Set_Val it can change value
        SETB    FIRST_SET                   ;Set flag to bypass 400 ms delay on
                                            ; 1st incrementation/decrementation
                                            ; and to start watering in Test
                                            ; Sprinklers menu
                ; Stop flashing during the incrementation or decrementation
                ;  to make the display easier to read while the value changes.
                ;  Do this by disabling the Flash function of the HDSP-2111
                ;  (bit 3 of CWR) so that we don't have to remember which
                ;  characters were flashing when the flashing is re-enabled.
        ANL     DIS_CWR, #0F7h              ;Stop flashing
        AJMP    EButt

Men:    CLR     SET_ACTIVE              ;All other cases: Set_Val must do
                                        ; nothing
        ORL     DIS_CWR, #08h           ;Resume flashing in case it was stopped
                                        ; during incrementation or decr.

                ; Menu button (PB1) is depressed
        CJNE    BUTTON_POS, #MENU, Sbmen ;PB1 is depressed

        JB      PROG_ACTIVE, Sbmen      ;If a Program is active, can't enter
                                        ; Menus

        MOV     SUBMENU_STA, #0         ;Reset Submenu

        INC     MENU_STA                ;Go to next Menu
        CJNE    MENU_STA, #MENU_MOD, MSWS ;MENU_STA is incremented modulo
        MOV     MENU_STA, #0              ; MENU_MOD
                ; Previous value of MENU_STA was 7, i.e. Test Sprinklers Menu.
                ;  Since we just exited it, we must stop watering.
        SETB    VOUT                    ;Stop watering

MSWS:   CJNE    MENU_STA, #6, MTS1      ;If Set Weekly Schedule menu, Schedule
        MOV     ZONE_NUM, #NONE_SCH     ; All command will start w/ enable PR1,
                                        ; disable PR2 (after incrementation)

MTS1:   CJNE    MENU_STA, #7, Sbmen     ;If Test Sprinklers menu,
        SJMP    ITS                     ; initialize Test Sprinklers function

                ; Submenu button (PB2) is depressed
Sbmen:  CJNE    BUTTON_POS, #SUBMENU, ESbm ;PB2 is depressed
        CLR     INIT_ZONE               ;Go to next zone (for "Skip Zone" cmd)

        INC     SUBMENU_STA             ;Go to next Submenu
        MOV     A, SUBMENU_STA
        CJNE    A, SUBMENU_MOD, MTS2    ;SUBMENU_STA is incremented modulo
        MOV     SUBMENU_STA, #0         ; SUBMENU_MOD

MTS2:   CJNE    MENU_STA, #7, ESbm      ;Special case of Test Sprinklers menu
                    ; Initialization of Test Sprinklers function
ITS:    CLR     FIRST_SET               ;Wait for PB3 to be depressed
        SETB    INIT_ZONE               ;Do not start watering and force
        SETB    END_ZONE                ; display of remaining duration to 00
        SETB    VOUT                    ;Stop watering

ESbm:           ; Stop button (PB4) is depressed: exit Menu (if in a menu)
                ;  and abort program (if any)
        CJNE    BUTTON_POS, #STOP, MPR  ;PB4 is depressed
        MOV     MENU_STA, #0            ;Exit Menu (return to Menu Idle)
        AJMP    TProg                   ;Abort program

                ; Manual Start
MPR:    CJNE    MENU_STA, #0, EButt     ;If in a Menu, manual start is disabled

                ; PB1 and PB3 buttons are depressed: manual start of PR1
MPR1:   CJNE    BUTTON_POS, #MST_PR1, MPR2  ;PB1 and PB3 are depressed
        AJMP    SPR1                    ;Go start PR1

                ; PB2 and PB3 buttons are depressed: manual start of PR2
MPR2:   CJNE    BUTTON_POS, #MST_PR2, EButt ;PB2 and PB3 are depressed
        AJMP    SPR2                    ;Go start PR2

                ; Schedule All command: done in the Menu section (label ScAll)

EButt:


;*****************************************************************************
;
; Menus
; -----
;
;   switch (MENU_STA)
;
;   case 0:  Menu Idle
;            Display Routine =
;               - if no Program active and no humidity: Dis_Clk
;               - if no Program active and humidity:
;                       * Dis_Clk on odd seconds
;                       * in line code calling Hum_Msg on even seconds
;               - if Program #1 or #2 active: Dis_Dur
;
;   case 1:  Set Clock
;            Display Routine = Dis_Clk
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Days
;       case 1:  Set Hours
;       case 2:  Set Minutes
;
;   case 2:  Set PR1 Start Times
;            Display Routine = Dis_Stt
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Hours of PR1, Start Time A
;       case 1:  Set Minutes of PR1, Start Time A
;       case 2:  Set Hours of PR1, Start Time B
;       case 3:  Set Minutes of PR1, Start Time B
;
;   case 3:  Set PR1 Durations
;            Display Routine = Dis_Dur
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Duration of ZONE1 in PR1
;       case 1:  Set Duration of ZONE2 in PR1
;       case 2:  Set Duration of ZONE3 in PR1
;       case 3:  Set Duration of ZONE4 in PR1
;       case 4:  Set Duration of ZONE5 in PR1
;       case 5:  Set Duration of ZONE6 in PR1
;       case 6:  Set Duration of ZONE7 in PR1
;       case 7:  Set Duration of ZONE8 in PR1
;
;   case 4:  Set PR2 Start Times
;            Display Routine = Dis_Stt
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Hours of PR2, Start Time A
;       case 1:  Set Minutes of PR2, Start Time A
;       case 2:  Set Hours of PR2, Start Time B
;       case 3:  Set Minutes of PR2, Start Time B
;
;   case 5:  Set PR2 Durations
;            Display Routine = Dis_Dur
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Duration of ZONE1 in PR2
;       case 1:  Set Duration of ZONE2 in PR2
;       case 2:  Set Duration of ZONE3 in PR2
;       case 3:  Set Duration of ZONE4 in PR2
;       case 4:  Set Duration of ZONE5 in PR2
;       case 5:  Set Duration of ZONE6 in PR2
;       case 6:  Set Duration of ZONE7 in PR2
;       case 7:  Set Duration of ZONE8 in PR2
;
;   case 6:  Set Weekly Schedule
;            Display Routine = Dis_Sch
;
;       switch (SUBMENU_STA)
;
;       case 0:  Set Monday Schedule
;       case 1:  Set Tuesday Schedule
;       case 2:  Set Wednesday Schedule
;       case 3:  Set Thursday Schedule
;       case 4:  Set Friday Schedule
;       case 5:  Set Saturday Schedule
;       case 6:  Set Sunday Schedule
;
;   case 7:  Test Sprinklers
;            Display Routine =
;               - if no Zone active (Submenu #0): Dis_Pr + in line code
;               - if Zone active (Submenu #1-#8): Dis_Dur
;
;       switch (SUBMENU_STA)
;
;       case 0:  Test ZONE1
;       case 1:  Test ZONE2
;       case 2:  Test ZONE3
;       case 3:  Test ZONE4
;       case 4:  Test ZONE5
;       case 5:  Test ZONE6
;       case 6:  Test ZONE7
;       case 7:  Test ZONE8
;
;*****************************************************************************

MMenu:
                    ; Always display a blank in 3rd char of Display
                    ;  (only overwritten by the humidity message and by
                    ;   Set PR1 / PR2 Start Times menus: "P1A..." to "P2B...")
        MOV     R7, #' '                ;Get blank character
        MOV     R0, #DIS_CH3            ;Get addr of 3rd character of Display
        ACALL   Out_Char                ;Output character
EBlk:

        CJNE    MENU_STA, #0, M1        ;Menu #0: Menu Idle

        ;-----------
        ; Menu Idle
        ;-----------

        JB      PROG_ACTIVE, EIdl       ;If a program is active, don't display
                                        ; Clock nor humidity message

                    ; If too much humidity, display the humidity message
                    ;  during even seconds (this will make this message
                    ;  alternate with the Clock every second)
        JB      HUMP, DClk              ;Not much humidity, no humidity msg
        JB      SECONDS.0, DClk         ;Odd seconds, display Clock

                ; Display humidity message
        MOV     R6, #1                  ;Initialize message index (1 rather
                                        ; than 0 because no "INC A" in Hum_Msg)
        MOV     R0, #DIS_CH1            ;Start with 1st character of Display
WHum:
        MOV     A, R6                   ;Get message index
        ACALL   Hum_Msg                 ;Get character in message
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        ACALL   Out_Char                ;Output character to Display
        INC     R6                      ;Point to next character in message
        INC     R0                      ;Point to next character of Display
        CJNE    R0, #DIS_CH8+1, WHum    ;Loop until 8 characters are written
        SJMP    EIdl

DClk:   ACALL   Dis_Clk                 ;Display Clock

EIdl:   AJMP    NoSet                   ;Stop Flashing and clear "Step-by-step"


M1:     CJNE    MENU_STA, #1, M2        ;Menu #1

        ;-----------
        ; Set Clock
        ;-----------

        ACALL   Dis_Clk                 ;Display Clock
        MOV     SUBMENU_MOD, #3         ;3 submenus in Menu #1

        CJNE    SUBMENU_STA, #0, SM11   ;Submenu #0

                ; Set Days of Clock
        MOV     FLASH_STA, #FLASH_1_2   ;Flash Days
        MOV     R0, #DAYS               ;Pass address of Days to Set_Val
        MOV     R7, #7                  ;Pass Modulo Days to Set_Val
        AJMP    SetV                    ;Set value and exit Menus

SM11:   CJNE    SUBMENU_STA, #1, SM12   ;Submenu #1

                ; Set Hours of Clock
        MOV     FLASH_STA, #FLASH_4_5   ;Flash Hours
        MOV     R0, #HOURS              ;Pass address of Hours to Set_Val
        MOV     R7, #24h                ;Pass Modulo Hours to Set_Val
        AJMP    SetV                    ;Set value and exit Menus

SM12:                                   ;Submenu #2
                ; Set Minutes and clear Seconds of Clock
        MOV     FLASH_STA, #FLASH_7_8   ;Flash Minutes
        MOV     R0, #MINUTES            ;Pass address of Minutes to Set_Val
        MOV     R7, #60h                ;Pass Modulo Minutes to Set_Val
        MOV     SECONDS, #0             ;Also clear Seconds
        AJMP    SetV                    ;Set value and exit Menus


M2:     CJNE    MENU_STA, #2, M3        ;Menu #2

        ;---------------------
        ; Set PR1 Start Times
        ;---------------------

        MOV     R0, #PR1A_HOUR          ;Get base address of PR1 start times
        MOV     PR_DIS, #PR1_DIS        ;To tell Dis_Stt it's PR1
        SJMP    SSta


M3:     CJNE    MENU_STA, #3, M4        ;Menu #3

        ;---------------------------
        ; Set Zone Durations in PR1
        ;---------------------------

                    ; PR_DIS (R5) still contains PR1_DIS from
                    ;  "PR1 Start Times" (to tell Dis_Dur it's PR1)
        MOV     A, #PR1_ZONE1           ;Base address (Duration of ZONE1
        SJMP    SDur                    ; in PR1)


M4:     CJNE    MENU_STA, #4, M5        ;Menu #4

        ;---------------------
        ; Set PR2 Start Times
        ;---------------------


        MOV     R0, #PR2A_HOUR          ;Get base address of PR2 start times
        MOV     PR_DIS, #PR2_DIS        ;To tell Dis_Stt it's PR2

        ; Set PR1 and PR2 Start Times
        ;----------------------------

;-----------------------------------------------------------------------------
;
; Notes:
;
;    1) To compute R1, input parameter of Dis_Stt, we mask bit 0 of the
;       submenu and we add it to the base address PR1A_HOUR or PR2A_HOUR
;       (in R0) so that Dis_Stt can display Start Time A when Submenu = 0
;       or 1 (R1 = PR1A_HOUR or PR2A_HOUR) and Start Time B when Submenu
;       = 2 or 3 (R1 = PR1B_HOUR or PR2B_HOUR).
;       This works because PR1B_HOUR is 2 bytes away from PR1A_HOUR and
;       PR2B_HOUR is 2 bytes away from PR2A_HOUR.
;
;    2) Transparent 'B' Start Times: see paragraph "Choice of a solution
;       to make 'B' Start Times transparent" at the beginning of the file.
;
;       Note on the implementation: we need to call Set_Val a first time to
;       set the 'A' Start Times and then a second time to set the 'B' Start
;       Times. We need to do a few tricks to make sure that the 'B' value is
;       incremented (idem decremented) when and only when the 'A' value is
;       incremented, and that there is no delay between these two operations.
;
;       To eliminate the delay between the incrementations of the 'A' and 'B'
;       values, we set flag FIRST_SET before calling Set_Val the second
;       time. It's not the primary purpose of this flag, but that does the
;       trick. But since the delay is eliminated, the 'B' value would be
;       incremented continuously (i.e. at every Background loop) if we did
;       not make sure that 'B' is incremented only if 'A' was incremented.
;       This is done by testing flag DLY_ACTIVE, which is clear if and
;       only if the value passed to it has been incremented (note that flag
;       SET_ACTIVE cannot change between the two calls to Set_Val, so we
;       don't have to worry about this one).
;
;-----------------------------------------------------------------------------

SSta:   MOV     SUBMENU_MOD, #4         ;4 submenus in Menus #2 and #4

        MOV     A, SUBMENU_STA          ;Copy submenu into Acc to mask bit
        CLR     ACC.0                   ;Mask bit 0
        ADD     A, R0                   ;Add base addr PR1A_HOURN or PR2A_HOUR
        MOV     R1, A                   ;Pass address of Hours of Start Time A
                                        ; or B to Dis_Stt
        ACALL   Dis_Stt                 ;Display Start time (increments R1
                                        ; so it now point to Minutes)
        MOV     R_0, R1                 ;Transfer address of Minutes to R0 for
                                        ; use by Set_Val

        MOV     A, SUBMENU_STA          ;Copy submenu into Acc for bit testing
        JB      ACC.0, SMin             ;If Submenu #1 or #3, set minutes

                ; Submenu #0 or #2
                ; Set Hours of Start Time A or B
        MOV     FLASH_STA, #FLASH_4_5   ;Flash Hours of Start Time
        DEC     R0                      ;Point to Hours
        MOV     R7, #24h                ;Pass Modulo Hours to Set_Val
        ACALL   Set_Val                 ;Set value

                ; Transparent 'B' Start Times: if setting Start Time A,
                ;  duplicate change in Start Time B
        MOV     A, SUBMENU_STA          ;Copy submenu into Acc for bit testing
        JB      ACC.1, EMen             ;If setting Start Time B, exit Menus
        INC     R0                      ;Point to Start Time B (2 bytes away
        INC     R0                      ; from corresponding Start Time A)
        AJMP    SetVB                   ;Set value (taking 400 ms delay into
                                        ; account) and exit Menus

                ; Submenu #1 or #3
                ; Set Minutes of Start Time A or B
SMin:   MOV     FLASH_STA, #FLASH_7_8   ;Flash Minutes of Start Time
        MOV     R7, #60h                ;Pass Modulo Minutes to Set_Val
        ACALL   Set_Val                 ;Set value

                ; Transparent 'B' Start Times: if setting Start Time A,
                ;  duplicate change in Start Time B
        MOV     A, SUBMENU_STA          ;Copy submenu into Acc for bit testing
        JB      ACC.1, EMen             ;If setting Start Time B, exit Menus
        INC     R0                      ;Point to Start Time B (2 bytes away
        INC     R0                      ; from corresponding Start Time A)
                            ; Set 'B' value, taking 400 ms delay into account
SetVB:  JB      DLY_ACTIVE, EMen        ;Don't set value if 400 ms delay active
                                        ; (i.e. only set 'B' if 'A' was set)
        SETB    FIRST_SET               ;No delay between the settings of 'A'
                                        ; and 'B'
        AJMP    SetV                    ;Set value and exit Menus


M5:     CJNE    MENU_STA, #5, M6        ;Menu #5

        ;---------------------------
        ; Set Zone Durations in PR2
        ;---------------------------

                    ; PR_DIS (R5) still contains PR2_DIS from
                    ;  "PR2 Start Times" (to tell Dis_Dur it's PR2)
        MOV     A, #PR2_ZONE1           ;Base address (Duration of ZONE1
                                        ; in PR2)

        ; Set Durations for PR1 and PR2
        ;------------------------------

SDur:   MOV     SUBMENU_MOD, #N_ZONE    ;8 submenus in Menus #3 and #5
                                        ; (1 for each zone)
                    ; To pass address of duration to Set_Val, add SUBMENU_STA
                    ;  (= zone number) to the base address #PR1_ZONE1
                    ;  or #PR2_ZONE1
        ADD     A, SUBMENU_STA          ;Add Submenu = number of the zone
        MOV     R1, A                   ;Address of Duration for Dis_Dur
        MOV     ZONE_NUM, SUBMENU_STA   ;Zone number for Dis_Dur
        ACALL   Dis_Dur                 ;Display Duration (doesn't smash R1)
        MOV     FLASH_STA, #FLASH_7_8   ;Flash Minutes (Duration)
        MOV     R_0, R1                 ;Pass address of Duration to Set_Val
                    ; For all durations, max value = 99, so modulo = 100
                    ;  (R7 = 00h is equivalent to 100h, i.e. 100 in BCD format)
        MOV     R7, #00h                ;Pass Modulo to Set_Val
        AJMP    SetV                    ;Set value and exit Menus


M6:     CJNE    MENU_STA, #6, M7        ;Menu #6

        ;------------------------
        ;   Schedule All command
        ;------------------------

                ; PB1 and PB2 buttons are depressed: Schedule All
ScAll:  CJNE    BUTTON_POS, #MST_PR1, WkSc ;PB1 and PB2 are depressed
        MOV     A, ZONE_NUM             ;Schedule value
        ACALL   Sch_All                 ;Write value in the 7 days
                    ; Get next schedule value
        JB      SCH_ALL_RDY, WkSc       ;Make sure we increment once only,
        SETB    SCH_ALL_RDY             ; every time PB1 and PB2 are depressed
        MOV     R0, #ZONE_NUM           ;Pass address of Schedule value to
                                        ; Inc_BCD
        MOV     R7, #SCH_MOD            ;Pass Modulo to Inc_BCD
        ACALL   Inc_BCD                 ;Increment Schedule value
                ; Note: the Schedule value is:
                ;   - NONE_SCH = 0 to disable PR1 and PR2
                ;   - PR1_SCH  = 1 to enable PR1 and disable PR2
                ;   - PR2_SCH  = 2 to enable PR2 and disable PR1
                ;   - BOTH_SCH = 3 to enable PR1 and PR2

        ;-----------------------
        ;   Set Weekly Schedule
        ;-----------------------

WkSc:   MOV     SUBMENU_MOD, #7         ;7 submenus in Menu #6

                    ; To pass address of schedule to Set_Val, add SUBMENU_STA
                    ;  (= day of the week) to the base address #SCH_MON
        MOV     A, #SCH_MON             ;Base address (Monday Schedule)
        ADD     A, SUBMENU_STA          ;Add Submenu = day of the week
        MOV     R1, A                   ;Address of Schedule for Dis_Sch
        ACALL   Dis_Sch                 ;Display Schedule for that day
                                        ; (doesn't smash R1)
        MOV     FLASH_STA, #FLASH_4_8   ;Flash chars 4, 5, 7 & 8 (Schedule)
        MOV     R_0, R1                 ;Pass address of Schedule to Set_Val
                    ; For all schedules, max value = 3, so modulo = 4
        MOV     R7, #04h                ;Pass Modulo to Set_Val
                                ; Fall through to Set value and exit Menus

SetV:   ACALL   Set_Val                 ;Set value
EMen:   AJMP    EMenu                   ;Exit Menus


M7:             ; Menu #7

        ;-----------------
        ; Test Sprinklers
        ;-----------------

;-----------------------------------------------------------------------------
;
; Notes:
;
;  1) The test duration is limited to 60 minutes in each zone so that in
;     case the user forgets to stop the test, watering does not continue
;     indefinitely. When the 60 minutes duration is exceeded, watering
;     stops and the Watering Controller stays in this state until the
;     user either tests the next zone or exits the test. This mechanism
;     is implemented with the END_ZONE flag in the Ctrl_Dur subroutine.
;     It is set when the duration is over. Then this flag is used to
;     skip the time comparison, which prevents restarting watering in
;     this zone for another 60 minutes, since MINS is reset to 0 after
;     it reached 60.
;
;  2) When entering the Test Sprinklers menu (via the Menu button, PB1)
;     and when selecting the zone to test (via the Submenu button, PB2),
;     an initialization is performed so that watering is not started
;     immediately. Watering will be started by depressing the Set button
;     (PB3). This permits to select a zone without starting watering
;     (even briefly) in any of the previous zones.
;
;     The initialization of the Test Sprinklers function is performed in
;     section "Debounce the buttons" (see Special case of Test Sprinklers"
;     menu"). Note that we had to play a few tricks because of the RAM
;     shortage: this is why we use some flags which were not initially
;     intended for this purpose. We can use FIRST_SET because this flag
;     is set when PB3 is depressed. It is cleared when routine Set_Val
;     is called, but this is never the case in the Test Sprinklers menu.
;     We use flags INIT_ZONE and END_ZONE to force the same state at
;     initialization as when the duration is over, which has the desired
;     effects: a) no watering takes place, b) the remaining duration
;     displayed is 00.
;
;      END_ZONE | INIT_ZONE | State
;     ----------+-----------+------------
;         0     |     0     | Start watering
;         0     |     1     | Watering
;         1     |     0     | ---
;         1     |     1     | No watering (watering not
;               |           |  started or duration over)
;
;
;-----------------------------------------------------------------------------

        MOV     SUBMENU_MOD, #N_ZONE    ;8 submenus in Menu #7 (1 per zone)

        MOV     PR_DIS, #TST_DIS        ;To print "TS"
        MOV     ZONE_NUM, SUBMENU_STA   ;Get Zone number = Submenu
        JBC     FIRST_SET, WTS          ;PB3 depressed, start watering (clear
        SJMP    DTS                     ; flag to initialize the zone once)
                ; Start watering
WTS:    CLR     INIT_ZONE               ;Initialize the zone

DTS:    MOV     R7, #60h                ;Set a fixed duration of 60 minutes
        ACALL   Ctrl_Dur                ;Display "TS", Zone number & duration,
                                        ; and control watering as per flags
                                        ; END_ZONE and INIT_ZONE

NoSet:      ; Menu Idle or Test Sprinklers menu
        ANL     FLASH_STA, #NO_FLASH_B6 ;No flashing (except for the 6th
                                        ; character if an outage has occurred)
EMenu:


;*****************************************************************************
;
; Execute Watering Program
; ------------------------
;
; This is where we control the watering zones, as per programs PR1 and PR2.
;
;
; Notes:
;
;  1) There are various ways of resolving a conflict between Menus and
;     Programs. We could either:
;       1 - take no special measure;
;       2 - abort the currently active program when entering a menu;
;       3 - leave the program run when entering a menu but do not start
;           a program when already in a menu;
;       4 - prevent access to menus if a program is active.
;     We have chosen solution 4: it is not possible to enter a menu if a
;     program is running. The program must be aborted first using the
;     "Manual Stop" command (depressing PB4). Furthermore, a program will
;     not start if we are already in a menu when it is supposed start.
;     With this solution, a program can never be active while we are in a
;     menu. This is the neater solution, at the expense of a minor
;     restriction for the user (entering menus is not a frequent operation
;     and is unlikely to be done while watering is taking place).
;
;  2) When a program is active, we ignore the start times of the other
;     program (or Start Time B if it was Start Time A and vice-versa).
;     This is accordance with the policy described in paragraph "Program
;     overlapping" at the beginning of the file.
;
;  3) Testing that seconds are equal to zero as well as testing that
;     hours and minutes match a start time make it possible to abort
;     a program (i.e. stop it manually by depressing PB4) at any time.
;     If seconds were not tested, it would not be possible to abort a
;     program during one minute after its start time because during that
;     minute the start time would be still met, and so the program would
;     be restarted immediately. Note that actually, by testing seconds
;     but not hundredths of seconds, it is still not possible to abort a
;     program during the first second, but this is so brief that it's
;     hardly noticeable.
;
;  4) The accuracy of the durations in the case of a Manual Start is only
;     1 minute, i.e. if for example PR1_ZONE1 = 10, then if Program #1 is
;     started manually, Zone #1 will be active for any time between 9 and
;     10 minutes. This is because the minutes for duration measurement (MINS)
;     are synchronized with the minutes of the Clock (MINUTES), and since at
;     the precise moment when the user starts the program manually, the
;     Clock may be at any time between two minutes, the first time MINS is
;     incremented may occur at any time between 0 and 1 minute. It would
;     be possible to fix this problem by letting MINS have its own seconds
;     (and hundredths of seconds!) but this is really not worth the effort...
;     (The problem does not exist for an automatic start since by nature
;     the start instant always occurs at the beginning of a minute of
;     the Clock - even if seconds were not tested against zero).
;
;  5) There is no conflict with the use of PR_DIS and ZONE_NUM by the
;     "Set PR1 / PR2 Start Times" menus because a Program cannot run
;     when a menu is active.
;
;*****************************************************************************

Pr_Act:
        JB      PROG_ACTIVE, BProg  ;If a program is already active, don't
                                    ; start another one
        CJNE    MENU_STA, #0, EProg ;Don't start a program if a menu is active
        JNB     HUMP, EProg         ;If too much humidity, don't start program
                                    ; (but humidity does not stop a program
                                    ; that has already started)
        MOV     A, SECONDS          ;If seconds not equal to zero, don't start
        JNZ     EProg               ; program (make it possible to abort prog
                                    ; during the first minute)

        ;--------------------------------
        ; Check Start Times and Schedule
        ;--------------------------------

;-----------------------------------------------------------------------------
;
; Note: first we'll see if the current time matches one of the four possible
;       start times (Start Times A and B of Programs #1 and #2), then if a
;       match if found we check the corresponding schedule.
;
;       The search for a match on one of the possible start times is based on
;       the fact that these times are contiguous in memory. We start with the
;       base address PR1A_HOUR. If it matches the hours of the Clock, we then
;       compare the minutes, incrementing the pointer (R0) to get PR1A_MINU.
;       If the minutes also match, we exit the loop to check the schedule and
;       to start the program if the schedule allows it. Else we check the next
;       starting time by looping back after incrementing the pointer (label
;       NStt), which will give us PR1B_HOUR. If the hours do no match, we
;       increment the pointer before branching to label NStt so that the
;       pointer gets incremented twice. So no matter whether there is a match
;       on the hours, whenever we loop back we point to the hours of the next
;       start time. Thus we test Start Time A of PR1, Start Time B of PR1,
;       Start Time A of PR2 and  Start Time B of PR2 in succession.
;
;       If the schedule does not allow a start time, we must loop back to
;       check the following start times because PR2 may have the same
;       start time as PR1 and unlike PR1 be enabled on that day.
;
;       When a match on a start time is found, we must determine if it was
;       on PR1 (i.e. Start Time A or B of PR1) or PR2 (i.e. Start Time A or
;       B of PR2). To do this, we use R6 as a flag while we scan the start
;       times: we initialize it to represent PR1 and we change it to represent
;       PR2 when we have scanned the start times of PR1 and we are prepared
;       to scan the start times of PR2. While we're at it, we put useful
;       information in R6 (the mask to check the schedule) rather than just
;       0 for PR1 and 1 for PR2.
;
;-----------------------------------------------------------------------------

Pr_Stt:
        MOV     R0, #PR1A_HOUR      ;Start with Start Time A of PR1
        MOV     R6, #PR1_SCH_MSK    ;Mask to check PR1 Schedule

LStt:   MOV     A, HOURS            ;Get Hours of the Clock
        CLR     C                   ;Clear Carry before subtracting
        SUBB    A, @R0              ;Subtract Hours of Start Time
        JZ      MStt                ;If hours match, check minutes
                        ; Hours do not match
        INC     R0                  ;Increment pointer (will be incremented
                                    ; again before looping back to LStt)
        SJMP    NStt                ;Go to next Start Time

MStt:   INC     R0                  ;Point to Minutes
        MOV     A, MINUTES          ;Get Minutes of the Clock
        CLR     C                   ;Clear Carry before subtracting
        SUBB    A, @R0              ;Subtract Minutes of Start Time
        JZ      SStt                ;If minutes match, go check schedule

                        ; Go to next Start Time
NStt:   INC     R0                  ;Increment pointer (now points to the hours
                                    ; of the next start time)
        CJNE    R0, #PR2A_HOUR, OStt ;Check if we enter the PR2 range
        MOV     R6, #PR2_SCH_MSK    ;If so, get mask to check PR2 Schedule
OStt:   CJNE    R0, #PR2B_HOUR+2, LStt ;If not done w/ 4 Start Times, continue
        SJMP    EProg               ;No match found, no program started

        ; Check Weekly Schedule
SStt:                   ; Get pointer to Weekly Schedule
        MOV     A, #SCH_MON         ;Base address (Monday Schedule)
        ADD     A, DAYS             ;Add day of the week
        MOV     R1, A               ;Copy address of Schedule into R1 for
                                    ; indirect addressing
        MOV     A, @R1              ;Get schedule
        ANL     A, R6               ;Mask for PR1 or PR2 as appropriate
        JZ      NStt                ;If program disabled, check next start time

                        ; Watering Program enabled, check if PR1 or PR2
        CJNE    R6, #PR1_SCH_MSK, SPR2 ;If match on PR2, branch

        ; Entry point for manual start of PR1
SPR1:   MOV     DURA_ACTIVE, #PR1_ZONE1 ;Point to duration of ZONE1 in PR1
        MOV     PR_DIS, #PR1_DIS    ;Tell Dis_Dur it's PR1
        SJMP    IProg               ;Don't check PR2 start times since
                                    ; PR1 is active

        ; Entry point for manual start of PR2
SPR2:   MOV     DURA_ACTIVE, #PR2_ZONE1 ;Point to duration of ZONE1 in PR2
        MOV     PR_DIS, #PR2_DIS    ;Tell Dis_Dur it's PR2

        ;----------------------------
        ; Watering Program Execution
        ;----------------------------

        ; Initialize program - begin with Zone #1
IProg:  MOV     ZONE_NUM, #0        ;Start with Zone #1

        ; Loop back here for next zone
NProg:  MOV     R0, DURA_ACTIVE     ;Transfer address of duration to R0
                                    ; for indirect addressing
        MOV     A, @R0              ;If duration is zero, make sure we
        JZ      NAsp                ; skip it w/o activating zone
        CLR     INIT_ZONE           ;To perform initialization of zone
        SJMP    DProg               ; in Ctrl_Dur

BProg:  JNB     INIT_ZONE, NAsp     ;If command to skip zone, go to next zone

DProg:  MOV     R0, DURA_ACTIVE     ;Transfer address of duration to R0
                                    ; for indirect addressing
        MOV     R_7, @R0            ;Get duration
        ACALL   Ctrl_Dur            ;Start watering and check duration
        JNC     EProg               ;If duration not over, exit (do
                                    ; Background loop), else go to next zone

        ; Next Zone
NAsp:   INC     DURA_ACTIVE         ;Point to duration of next zone
        INC     ZONE_NUM            ;Go to next zone
        MOV     A, ZONE_NUM         ;To compare ZONE_NUM in CJNE instruction
        CJNE    A, #N_ZONE, NProg   ;If not done w/ the 8 zones, do next zone,
                                    ; else terminate program

        ; Terminate Program
TProg:  SETB    VOUT                ;Stop watering
        MOV     DURA_ACTIVE, #0     ;Program is no longer active

EProg:


;*****************************************************************************
;
; Output Flashing Status to Display
; ---------------------------------
;
; Here we output the Flashing Status out of the Flashing Status image
;  FLASH_STA to the 8-Character Display.
;
; Note that we have to keep an image of the Flashing Status so that we can
;  output the Flashing Status of 8 characters at the same time, because the
;  the Flashing Status of each character is represented by a bit, so to
;  modify it directly in the Display would require that we are able to
;  read the Flashing Status out of the Display, and our hardware does not
;  provide the capability to read data out of the Display.
;
; See description of Out_Char subroutine, in particular for the address of
;  a character on the Display (FL/, A3, A2, A1 and A0).
;
; Note: FLASH_STA is rotated right 8 times, so at the end
;       of this section, its initial value is restored.
;*****************************************************************************

Out_Fla:
        MOV     R0, #DIS_FL1        ;Address of Flashing status of 1st char

LOut_Fl:                ; N.B. We'll start with bit 0 of FLASH_STA =
                        ;  Flashing status of 1st character
        MOV     R7, FLASH_STA       ;Pass Flashing status to Out_Char
                                    ; (only bit 0 is taken into consideration)
        ACALL   Out_Char            ;Output Flashing status
                        ; Get ready for next character
        MOV     A, FLASH_STA        ;Transfer FLASH_STA to Acc to rotate it
        RR      A                   ;Bit 0 = Flashing status of next character
        MOV     FLASH_STA, A        ;Transfer back to FLASH_STA
        INC     R0                  ;Address of Flashing status of next char
        CJNE    R0, #DIS_FL8+1, LOut_Fl ;Loop until Flashing status of
                                        ; 8 characters are output


;*****************************************************************************
;
; Monitor Mains Power (Power Down Feature)
; ----------------------------------------
;
; If loss of mains power is detected (via the Mains Power Input, MPIN),
;  the Display is blanked by setting the brightness to the minimum.
;  This permits the backup battery to last much longer, since the Display
;  draws about 100 mA (because we set the brightness to 53 %), whereas
;  the 87C750 and associated logic circuitry (74HC138 and 74HC164) together
;  will draw no more than about 10 mA. (Note: the 87C750 will draw less
;  than 11 mA with a 12 MHz clock and we use a 6 MHz clock).
;
; The Mains Power Input (MPIN) is connected to the input of the 5 V
;  regulator (via proper voltage reduction and protection), so that this
;  input will be low if the mains power is off (mains outage) although
;  the backup power is present.
;
; Since the Sprinkler valves are powered by a distinct power supply
;  (24V AC) which is not backed up, the Watering Controller will not
;  operate during an outage. The purpose of the backup battery is to make
;  sure that the Watering Controller will be operational as soon as the
;  outage is over, since it allows the Clock to continue working internally
;  and since it preserves the programming information (program start times,
;  durations, weekly schedule).
;
; When mains power is restored, we make the hours / minutes separator
;  (colon character, i.e. 6th character of the Display) flash to make it
;  visible that an outage has occurred. (Actually, we set the flashing when
;  the mains power is off, but since the Display is blank at this time,
;  the flashing will be visible when the mains power is restored).
;
; This flashing is removed as soon as any push-button is depressed
;  (indicating a user intervention).
;
; We change bits 2-0 of Control Word Register according to the state of
;  the Mains Power Input (MPIN):
;
; Mains power on  => normal brightness:  D2-D0 = 010 (53 % brightness)
; Mains power off => minimum brightness: D2-D0 = 111 ( 0 % brightness)
;
;*****************************************************************************

Mon_MPI:
        JB      MPIN, MPOn          ;If mains power is present, branch

                ; Mains power is off
        ORL     DIS_CWR, #07h       ;Set brightness to minimum
        SETB    FLASH_STA.5         ;Flash 6th character of Display
        SJMP    Out_CWR             ;Output CWR to the Display

MPOn:           ; Mains power is on
        ANL     DIS_CWR, #0FAh      ;Set brightness to normal
                    ; Fall through to output CWR to the Display


;*****************************************************************************
;
; Output Control Word Register to Display
; ---------------------------------------
;
; Outputs the image of the Control Word Register (DIS_CWR) to the Display
;  so that a new value of DIS_CWR can take effect.
;
;*****************************************************************************

Out_CWR:
        MOV     R7, DIS_CWR         ;Data for Control Word Register
        MOV     R0, #CWR_ADDR       ;Address of Control Word Register
        ACALL   Out_Char            ;Write to Control Word Register


;*****************************************************************************
;
; End of Background
; -----------------
;
; Call Clock subroutine and jump to the top of the Background loop.
;
;*****************************************************************************

        ACALL   Clock               ;Execute Clock
        AJMP    Bkgnd               ;Jump to top of Background loop


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


;=============================================================================
;
;   SUBROUTINES
;
;=============================================================================

;*****************************************************************************
;
; Set Value Subroutine
; --------------------
;
; Lets the user set a value by incrementing or decrementing the value (always
;  in BCD format) every 400 ms as long as he depresses the Set push-button.
;  Also permits to clear the value.
;
; It is used by Menus #1 to #6.
;
; Note: the 400 ms delay must not be performed with a blind loop because of
;       an undesired side effect on the display: the selected value appears
;       to be incremented right after the Set push-button is released,
;       which makes it tricky to set the desired value. We must execute
;       the background loop continuously during the 400 ms delay in order
;       to keep the Display correctly updated.
;
;       The reason why the selected value appears to be incremented right
;       after the Set push-button is released when the delay is performed
;       in a blind loop, is because when the Set push-button is released,
;       we are typically within the 400 ms delay for the next value, and
;       since the loop is blind, we execute it completely and we increment
;       the value.
;
; Input parameters:
;   R0 = Address of data to set
;   R7 = Modulo value
;      Note: a value of 00h is equivalent to a modulo of 100h.
;
; Output:
;   Data pointed by R0 is modified (incremented, decremented or cleared)
;   Note: Data = 0 if rolled over modulo
;
; Side effects: Acc is modified
;*****************************************************************************

Set_Val:
        JNB     SET_ACTIVE, XSet    ;Don't change value if Set button not
                                    ; depressed
        JBC     FIRST_SET, SSet     ;No delay for the first incrementation/
                                    ; decrementation. Clear flag for next time
        JB      DLY_ACTIVE, WSet    ;Don't initialize delay if delay active

        MOV     HUNDTHS, #0         ;Initialize 400 ms delay
        SETB    DLY_ACTIVE          ;Indicate initialization is done

WSet:   MOV     A, HUNDTHS          ;Go round background loop w/o incrementing
        CJNE    A, #INC_DLY, XSet   ; value until 400 ms have elapsed

                ; 400 ms delay elapsed
                        ;Note: it is OK to test BUTTON_POS 400 ms after buttons
                        ; were depressed because as long as FIRST_SET is
                        ; set, any change in button position is ignored
SSet:   CJNE    BUTTON_POS, #CLRVAL, TSet ;If PB4 set as well as PB3,
        MOV     @R0, #0                   ; clear value
        SJMP    ESet

TSet:   JB      SET_DOWN, DSet      ;If PB2 depressed as well as PB3, decrement
        ACALL   Inc_BCD             ;Else increment value pointed by R0
        SJMP    ESet

DSet:   ACALL   Dec_BCD             ;Decrement value pointed by R0

ESet:   CLR     DLY_ACTIVE          ;Get ready for next 400 ms delay
XSet:   RET


;*****************************************************************************
;
; Increment BCD Subroutine
; ------------------------
;
; Increments a value in BCD format, taking into account a modulo value.
;
; Input parameters:
;   R0 = Address of data to increment
;   R7 (R_7 = 07h) = Modulo value
;      Note: a value of 00h is equivalent to a modulo of 100h.
;
; Output:
;   Data pointed by R0 is incremented
;   Note: Data = 0 if rolled over modulo
;   Value incremented in also available in Acc
;
; Side effects: Acc is modified
;*****************************************************************************

Inc_BCD:
        MOV     A, @R0              ;Get value to increment
        ADD     A, #1               ;Increment value (cannot use INC for DA A!)
        DA      A                   ;Decimal adjust (2 BCD digits)
        CJNE    A, R_7, IMod        ;If modulo, value = 0
        CLR     A
IMod:   MOV     @R0, A              ;Restore incremented value
        RET



;*****************************************************************************
;
; Decrement BCD Subroutine
; ------------------------
;
; Decrements a value in BCD format, taking into account a modulo value.
;
; Input parameters:
;   R0 = Address of data to decrement
;   R7 (R_7 = 07h) = Modulo value
;      Note: a value of 00h is equivalent to a modulo of 100h.
;
; Output:
;   Data pointed by R0 is decremented
;   Note: Data = modulo - 1 if rolled under 0
;   Value decremented in also available in Acc
;
; Side effects: Acc is modified
;*****************************************************************************

Dec_BCD:
        MOV     A, @R0              ;Get value to decrement
        JNZ     DecB                ;Just decrement if not rolled under zero
        MOV     A, R7               ;Get Modulo value: will be decremented
                                    ; also, so result will be modulo - 1
DecB:   ADD     A, #99h             ;Decrement value (cannot use DEC nor SUBB
                                    ; for DA A!)
        DA      A                   ;Decimal adjust (2 BCD digits)
        MOV     @R0, A              ;Restore decremented value
        RET



;*****************************************************************************
;
; Schedule All Subroutine
; -----------------------
;
; This routine writes a constant value into the schedules of the 7 days of
;  the week.
;
; Input parameters:
;   Acc = value to be written into the 7 schedules
;         -  NONE_SCH = 0: disable PR1 and PR2
;         -  PR1_SCH  = 1: enable PR1, disable PR2
;         -  PR2_SCH  = 2: enable PR2, disable PR1
;         -  BOTH_SCH = 3: enable PR1 and PR2
;
; Side effects: R0 is modified
;*****************************************************************************

Sch_All:
        MOV     R0, #SCH_MON            ;Start address of Schedule area
WSch:
        MOV     @R0, A                  ;Write data into one RAM location
        INC     R0                      ;Address of next schedule location
        CJNE    R0, #SCH_SUN+1, WSch    ;Loop until the 7 locations are done
        RET



;*****************************************************************************
;
; Control Duration Subroutine
; ---------------------------
;
; This routine starts watering in a zone and checks the watering duration.
; When the duration is over, it returns with the Carry bit set.
;
; This routine is called by the "Test Sprinklers" menu and by the "Execute
;  Watering Program" section.
;
;
; Note on the method to output the zone address:
;
;   To output the zone address on bits 0 to 2 of Port 1 without affecting
;   the other bits, we must use an ORL instruction to set the address bits
;   that are equal to 1, with bits 3-7 equal to 0 to have no effect on the
;   latter, and then we must use an ANL instruction to clear the address
;   bits that are equal to 0, with bits 3-7 equal to 1 to have no effect
;   on the latter.
;
;   Example:
;    Zone number = Zone address = 02h = 0000 0010 in ZONE_NUM
;    Assume initial value on P1 = D1h
;                                             P1          Acc
;         MOV     A, ZONE_NUM             1101 0001    0000 0010
;         ORL     P1, A                   1101 0011    0000 0010
;         ORL     A, #0F8h                1101 0011    1111 1110
;         ANL     P1, A                   1101 0010    1111 1110
;
;    Bits 0-2 of P1 are now equal to bits 0-2 of ZONE_NUM and
;    bits 3-7 of P1 are unchanged.
;
;   In order to have a really clean operation, we de-activate VOUT/
;   before changing the address, because in this process, the address
;   gets an intermediate value that may be neither the old value nor
;   the new value. Note however that for practical matters, this
;   precaution is not really necessary, because the spurious address
;   appears for such a short time that the effect will be completely
;   invisible (due to the valve mechanical inertia).
;
;
; Input parameters:
;   ZONE_NUM = Zone number (0 for ZONE1, 1 for ZONE2, etc.)
;   R7 = total duration
;   PR_DIS = PR1_DIS (= '1') for Program #1
;            PR2_DIS (= '2') for Program #2
;            TST_DIS (= 'S') for Test Sprinklers
;   INIT_ZONE = flag FALSE (= 0) if initialization not done
;                    TRUE (= 1) if initialization done
;   END_ZONE  = flag TRUE (= 1) when zone duration is over
;
; Output:
;   Carry bit is set if duration is over, cleared otherwise
;
; Side effects: Acc, R0, R1, R6 and R7 are modified
;*****************************************************************************

Ctrl_Dur:
        JB      INIT_ZONE, EDIni    ;If initialization done, skip it

DIni:       ; Initialization for the zone
        MOV     MINS, #0            ;Clear minutes to measure duration
            ; Output zone address and start watering
        SETB    VOUT                ;De-activate watering until zone address
                                    ; is stable
        MOV     A, ZONE_NUM         ;Move zone address to Acc
        ORL     P1, A               ;Write 1's of zone addr in bits 0-2 of P1
        ORL     A, #0F8h            ;Set bits 3-7 of Acc to 1 so that ANL
                                    ; has no effect on those bits
        ANL     P1, A               ;Write 0's of zone addr in bits 0-2 of P1
        CLR     VOUT                ;Start watering
        SETB    INIT_ZONE           ;Indicate initialization is done
        CLR     END_ZONE            ;Start afresh with duration
EDIni:                  ; End of initialization

            ; Check if duration is over
        CLR     A                   ;So that remaining duration is displayed
                                    ; as zero when duration is over
        JB      END_ZONE, DDis      ;If zone timed out, skip checking of
                                    ; duration (for "Test Sprinklers")
        MOV     A, R7               ;Get total duration
        CLR     C                   ;Clear Carry before subtracting
        SUBB    A, MINS             ;Subtract minutes from duration
                ; Since we subtract BCD numbers, we must perform a "decimal
                ;  adjust" by subtracting 6 if there was a borrow on the low
                ;  order digit (we can't use DA A with subtraction)
        JNB     AC, ESub            ;Auxiliary Carry indicates borrow for
                                    ; the low order digit -> subtract 6
                ; No need to clear the Carry for the 2nd subtraction because
                ;  1st number is always larger than 2nd in the 1st subtraction
        SUBB    A, #6               ;Performs decimal adjust
ESub:   JNZ     DDis                ;If duration not over, don't set END_ZONE
        SETB    END_ZONE            ;Duration is over, stop watering

        ; Display current Program and Zone numbers and remaining duration
                ; Acc contains the remaining duration in minutes
DDis:   MOV     R6, A               ;Move duration to R6 for indirect
                                    ; addressing by Dis_Dur
        MOV     R1, #R_6            ;Pass address of duration to Dis_Dur
        ACALL   Dis_Dur             ;Display Program number, Zone number
                                    ; and remaining duration
        JB      END_ZONE, Over      ;If duration is over, exit with Carry set
        CLR     C                   ;Duration not over, exit with Carry clear
        RET

Over:   ; Duration over, stop watering and exit with Carry set
        SETB    VOUT                ;Stop watering
        SETB    C                   ;Set Carry
        RET



;*****************************************************************************
;
; Display Clock Subroutine
; ------------------------
;
; This routine is called in Menu Idle (MENU_STA = 0) and in the
;  "Set Clock" Menu (MENU_STA = 1).
;
; Side effects: Acc, R0, R6 and R7 are modified
;*****************************************************************************

Dis_Clk:
                ; Display Day of the Week
        MOV     R6, DAYS                ;Get current day of the week
        ACALL   Dis_Day                 ;Print it on 1st & 2nd chars of Display

                ; Display Hours
        MOV     R0, #DIS_CH4            ;Tens of hours on 4th char of Display
        MOV     R6, HOURS               ;Get hours (2-digit BCD number)
        ACALL   BCD_ASC                 ;Convert and output to Display

                ; Display Hours/Minutes Separator = ':'
        MOV     R7, #':'                ;Get ':' character
        MOV     R0, #DIS_CH6            ;Get addr of 6th character of Display
        ACALL   Out_Char                ;Output character

                ; Display Minutes
        MOV     R0, #DIS_CH7            ;Tens of mins on 7th char of Display
        MOV     R6, MINUTES             ;Get minutes (2-digit BCD number)
        ACALL   BCD_ASC                 ;Convert and output to Display

        RET



;*****************************************************************************
;
; Display Program Start Times Subroutine
; --------------------------------------
;
; This routine is called in the "Set PR1 Start Times" (MENU_STA = 2) and
;  "Set PR2 Start Times" (MENU_STA = 4) Menus.
;
; Input parameters:
;   PR_DIS = PR1_DIS (= '1') for Program #1
;            PR2_DIS (= '2') for Program #2
;   R1 = Address of Hours of Start Time
;   SUBMENU_STA = 0 or 1 to display Start Time A
;                 2 or 3 to display Start Time B
;
; Side effects: Acc, R0, R6 and R7 are modified
;               R1 is incremented (so that it points to Minutes of Start Time)
;*****************************************************************************

Dis_Stt:
                ; Display "P1" or "P2"
        MOV     R0, #DIS_CH1            ;Print 'P' on 1st char of Display
        ACALL   Dis_Pr                  ;Display "P1" or "P2"

                ; Display 'A' or 'B'
        MOV     R7, #'A'                ;Get 'A' character
        MOV     A, SUBMENU_STA          ;Get Submenu
        JNB     ACC.1, Dis_AB           ;If Submenu #0 or #1, it's 'A'
        MOV     R7, #'B'                ;Else, Submenu #2 or #3, so it's 'B'
Dis_AB: MOV     R0, #DIS_CH3            ;Get addr of 3rd character of Display
        ACALL   Out_Char                ;Output character


                ; Display Hours of Start Time
        MOV     R_6, @R1                ;Get Hours
        MOV     R0, #DIS_CH4            ;Tens of hours on 4th char of Display
        ACALL   BCD_ASC                 ;Convert and copy into Display Buffer

                ; Display Hours/Minutes Separator = ':'
        MOV     R7, #':'                ;Get ':' character
        MOV     R0, #DIS_CH6            ;Get addr of 6th character of Display
        ACALL   Out_Char                ;Output character

                ; Display Minutes of Start Time
        INC     R1                      ;Point to Minutes
        MOV     R_6, @R1                ;Get Minutes
        MOV     R0, #DIS_CH7            ;Tens of mins on 7th char of Display
        ACALL   BCD_ASC                 ;Convert and copy into Display Buffer
        RET



;*****************************************************************************
;
; Display Watering Duration Subroutine
; ------------------------------------
;
; This routine is called:
; 1) in the "Set PR1 Durations" (MENU_STA = 3) and
;    "Set PR2 Durations" (MENU_STA = 5) Menus;
; 2) in the "Test Sprinklers" Menu (MENU_STA = 7);
; 3) when a Program is running.
;
; Note: there is no conflict between any of these three cases because
;       a Program cannot run when a menu is active.
;
; Input parameters:
;   PR_DIS = PR1_DIS (= '1') for Program #1
;            PR2_DIS (= '2') for Program #2
;            TST_DIS (= 'S') for Test Sprinklers
;   ZONE_NUM = Zone number (0 for ZONE1, 1 for ZONE2, etc.)
;   R1 = Address of Duration
;
; Side effects: Acc, R0, R6 and R7 are modified
;*****************************************************************************

Dis_Dur:
        MOV     R0, #DIS_CH1            ;Print 'P'/'T' on 1st char of Display
        ACALL   Dis_Pr                  ;Display "P1", "P2" or "TS"

                 ; Print 'Z' on 4th character of Display
        MOV     R7, #'Z'                ;Get 'Z' character
        MOV     R0, #DIS_CH4            ;Get addr of 4th character of Display
        ACALL   Out_Char                ;Output character

                ; Display Zone number
        MOV     A, ZONE_NUM             ;Get Zone number
        ADD     A, #31h                 ;Add 1 and convert to ASCII
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        MOV     R0, #DIS_CH5            ;Get addr of 5th character of Display
        ACALL   Out_Char                ;Output Zone number on 5th char

                ; Print blank between "An" and Duration
        MOV     R7, #' '                ;Get blank character
        MOV     R0, #DIS_CH6            ;Get addr of 6th character of Display
        ACALL   Out_Char                ;Output character

                ; Display Duration (number of minutes)
                ; Note: when Dis_Dur is called by Ctrl_Dur, R1 holds the
                ;  address of R6, so the following MOV moves data onto
                ;  itself. The MOV is necessary when Dis_Dur is called by
                ;  the "Set Durations" section (label SDur).
        MOV     R_6, @R1                ;Get Duration
        MOV     R0, #DIS_CH7            ;Tens of mins on 7th char of Display
        ACALL   BCD_ASC                 ;Convert and copy into Display Buffer
        RET



;*****************************************************************************
;
; Display Schedule Subroutine
; ---------------------------
;
; This routine is called in the "Set Weekly Schedule" (MENU_STA = 6).
;
; Input parameters:
;   SUBMENU_STA = scheduled day of the week
;   R1 = Address of Schedule
;
; Side effects: Acc, R0, R6 and R7 are modified
;*****************************************************************************

Dis_Sch:
                ; Display Day of the Week
        MOV     R6, SUBMENU_STA_B       ;Get scheduled day of the week
        ACALL   Dis_Day                 ;Print it on 1st & 2nd chars of Display

                ; Print blank between "P1" and "P2"
        MOV     R7, #' '                ;Get blank character
        MOV     R0, #DIS_CH6            ;Get addr of 6th character of Display
        ACALL   Out_Char                ;Output character

HPR1:           ; Check if PR1 is enabled
        MOV     A, @R1                  ;Copy schedule into Acc for bit testing
        JB      PR1_ENA, DPR1           ;If PR1 is enabled, print "P1"
                        ; Else print '-' on 4th and 5th characters of Display
        MOV     R7, #'-'                ;Get '-' character
        MOV     R0, #DIS_CH4            ;Get addr of 4th character of Display
        ACALL   Out_Char                ;Output character
        MOV     R0, #DIS_CH5            ;Get addr of 5th character of Display
        ACALL   Out_Char                ;Output character
        SJMP    HPR2

DPR1:           ; Print "P1" on 4th and 5th characters of Display
        MOV     R0, #DIS_CH4            ;Print 'P' on 4th char of Display
        MOV     PR_DIS, #PR1_DIS        ;Print '1' on 5th char of Display
        ACALL   Dis_Pr                  ;Display "P1"

HPR2:           ; Check if PR2 is enabled
        MOV     A, @R1                  ;Copy schedule into Acc for bit testing
        JB      PR2_ENA, DPR2           ;If PR2 is enabled, print "P2"
                        ; Else print '-' on 7th and 8th characters of Display
        MOV     R7, #'-'                ;Get '-' character
        MOV     R0, #DIS_CH7            ;Get addr of 7th character of Display
        ACALL   Out_Char                ;Output character
        MOV     R0, #DIS_CH8            ;Get addr of 8th character of Display
        ACALL   Out_Char                ;Output character
        RET

DPR2:           ; Print "P2" on 7th and 8th characters of Display
        MOV     R0, #DIS_CH7            ;Print 'P' on 7th char of Display
        MOV     PR_DIS, #PR2_DIS        ;Print '2' on 8th char of Display
        ACALL   Dis_Pr                  ;Display "P2"
        RET



;*****************************************************************************
;
; Display Day of the Week Subroutine
; ----------------------------------
;
; The day of the week is always displayed on the 1st and 2nd characters
;  of the Display.
;
; Input parameters:
;   R6 = Day of Week
;
; Side effects: Acc, R0 and R7 are modified
;*****************************************************************************

Dis_Day:
        MOV     A, R6                   ;Day of the week = index into table
        ACALL   Day_Ch1                 ;Get 1st Char of the Day of the Week
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        MOV     R0, #DIS_CH1            ;Get addr of 1st character of Display
        ACALL   Out_Char                ;Output 1st Char of the Day of the Week
        MOV     A, R6                   ;Day of the week = index into table
        ACALL   Day_Ch2                 ;Get 2nd Char of the Day of the Week
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        MOV     R0, #DIS_CH2            ;Get addr of 2nd character of Display
        ACALL   Out_Char                ;Output 2nd Char of the Day of the Week
        RET



;*****************************************************************************
;
; Humidity message Subroutine
; ---------------------------
;
; This routine is called with the position of the character in the Acc
; ('H' = 1st character, etc.) and returns the character of the humidity
; message at this position.
;
; Trick: this routine is called with Acc = 1 rather than 0 for the 1st
;  character, etc., so we don't need the "INC A" instruction.
;
; Side effects: Acc is modified
;*****************************************************************************

Hum_Msg:
        MOVC    A, @A+PC                ;Fetch value from table
        RET
        DB      'H'
        DB      'U'
        DB      'M'
        DB      'I'
        DB      'D'
        DB      ' '
        DB      ' '
        DB      ' '



;*****************************************************************************
;
; Get 1st Character of the Day of the Week Subroutine
; ---------------------------------------------------
;
; This routine is called with the day of the week in the Acc (Monday =  0,
;  etc.) and returns the first letter of the day of the week (in French).
;
; Side effects: Acc is modified
;*****************************************************************************

Day_Ch1:
        INC     A                       ;Table starts after RET instruction
        MOVC    A, @A+PC                ;Fetch value from table
        RET
                ;First letter of each day of the week in English
        DB      'M'                     ;Monday
        DB      'T'                     ;Tuesday
        DB      'W'                     ;Wednesday
        DB      'T'                     ;Thursday
        DB      'F'                     ;Friday
        DB      'S'                     ;Saturday
        DB      'S'                     ;Sunday



;*****************************************************************************
;
; Get 2nd Character of the Day of the Week Subroutine
; ---------------------------------------------------
;
; This routine is called with the day of the week in the Acc (Monday =  0,
;  etc.) and returns the second letter of the day of the week (in French).
;
; Side effects: Acc is modified
;*****************************************************************************

Day_Ch2:
        INC     A                       ;Table starts after RET instruction
        MOVC    A, @A+PC                ;Fetch value from table
        RET
                ;Second letter of each day of the week in English
        DB      'O'                     ;mOnday
        DB      'U'                     ;tUesday
        DB      'E'                     ;wEdnesday
        DB      'H'                     ;tHursday
        DB      'R'                     ;fRiday
        DB      'A'                     ;sAturday
        DB      'U'                     ;sUnday



;*****************************************************************************
;
; Display Program Number Subroutine
; ---------------------------------
;
; This routine displays "P1" for Program #1 (PR1), "P2" for Program #2 (PR2)
;  and "TS" for TeSt on the characters of the Display pointed to by R0
; (1st and 2nd characters if R0 = 0).
;
; Input parameters:
;   R0 = Address in the Display of the first of the two characters to be
;        printed
;        ('P' if Program number or 'T' if Test)
;        i.e. 0 = 1st char, 1 = 2nd char, etc.
;   PR_DIS = PR1_DIS (= '1') for Program #1
;            PR2_DIS (= '2') for Program #2
;            TST_DIS (= 'S') for Test Sprinklers
;
; Side effects: Acc, R0 and R7 are modified
;*****************************************************************************

Dis_Pr:
        MOV     R7, #'P'            ;Assume it's Prog, hence 1st char = 'P'
        MOV     A, PR_DIS           ;Get second char to be printed
        CJNE    A, #TST_DIS, DP2    ;If not Program, then it's Test
        MOV     R7, #'T'            ;1st char = 'T'
DP2:    ACALL   Out_Char            ;Output first char to be printed
        INC     R0                  ;Point to next char of Display
        MOV     R7, PR_DIS_B        ;Get 2nd char ('1', '2' or 'S')
        ACALL   Out_Char            ;Output second char to be printed
        RET



;*****************************************************************************
;
; BCD to ASCII Conversion Subroutine
; ----------------------------------
;
; This routine converts a 2-digit BCD number into 2 ASCII characters that
;  will be displayed on the 8-Character Display
;
; Input parameters:
;   R0 = Address of the 1st digit in the Display (tens)
;   R6 = BCD number
;
; Output:
;   Update Display at locations Ad and Ad+1, where Ad is the content of R0
;
; Side effects: Acc, R0 and R7 are modified
;*****************************************************************************

BCD_ASC:
        MOV     A, R6                   ;Get tens
        SWAP    A                       ;Move tens to low nibble
        ANL     A, #0Fh                 ;Clear high byte
        ORL     A, #30h                 ;Convert to ASCII
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        ACALL   Out_Char                ;Output character
        INC     R0                      ;Point to next char on Display
        MOV     A, R6                   ;Get units (already in low nibble)
        ANL     A, #0Fh                 ;Clear high byte
        ORL     A, #30h                 ;Convert to ASCII
        MOV     R7, A                   ;Transfer character to R7 for Out_Char
        ACALL   Out_Char                ;Output character
        RET



;*****************************************************************************
;
; Output Character to Display Subroutine
; --------------------------------------
;
; This routine outputs one character to the 8-Character Display.
; It is called with the address and ASCII code of the character.
; It can also be used to write the Flashing status of the character
; by using a different address.
; It is also used to write data to the Control Word Register.
;
;
; Address
; - - - -
;
; In this subroutine, R0 holds the "address of character" in the broad
;  sense, because this address consists of FL/, A3, A2, A1 and A0.
;  FL/ is treated as an address line, equivalent to A4 in our case,
;  since the A4 line of the Display is not used (tied to VDD). This
;  permits not only to write an ASCII code in any character of the
;  Display but also to write the Flashing status in any character and
;  to write to the Control Word Register.
;
;       FL/   A3    A2    A1    A0
;       bit4  bit3  bit2  bit1  bit0
;
; So the addresses can take the following values, with the following meanings:
;
;       08h = Flashing status, 1st character
;       09h = Flashing status, 2nd character
;       ...
;       0Fh = Flashing status, 8th character
;       10h = Control Word Register
;       18h (DIS_CH1) = ASCII code, 1st character
;       19h (DIS_CH2) = ASCII code, 2nd character
;       ...
;       1Fh (DIS_CH8) = ASCII code, 8th character
;
; To output the address on bits 0 to 4 of Port 3 without affecting the other
;  bits, we must use an ORL instruction to set the address bits that are
;  equal to 1, with bits 5-7 equal to 0 to have no effect on the latter,
;  and then we must use an ANL instruction to clear the address bits that
;  are equal to 0, with bits 5-7 equal to 1 to have no effect on the latter
;  (this is the same mechanism as we used to output the zone address in
;  the Ctrl_Dur subroutine).
;
; Example:
;  Address = 06h = 0000 0110 in R0
;  Assume initial value on P3 = 23h
;                                           P3          Acc
;       MOV     A, R0                   0010 0011    0000 0110
;       ORL     P3, A                   0010 0111    0000 0110
;       ORL     A, #0E0h                0010 0111    1110 0110
;       ANL     P3, A                   0010 0110    1110 0110
;
; Bits 0-4 of P3 are now equal to bits 0-4 of R0 and bits 5-7 of P3
;  are unchanged.
;
;
; Data transfer
; - - - - - - -
;
; The character is output bit by bit in serial form (bit 0 first).
;  An external serial to parallel shift register converts it to parallel
;  for the Display.
;
; We assume the data is transferred to the shift register on the rising edge
;  of the Serial Clock (actually, it would work with the falling edge also
;  because the data remains valid until after the falling edge).
;
;
;                       ---------------------------------------------
;  Serial Data  (DSD)     | Bit 0 | Bit 1 | ..... | ..... | Bit 7 |
;                       ---------------------------------------------
;
;                             --      --      --      --      --
;  Serial Clock (DSC)        |  |    |  |    |  |    |  |    |  |
;                       -----    ----    ----    ----    ----    ----
;
;
; Notes:
;
;    1) Since bit 7 is always zero in the data that we write, we shall
;       always access an ASCII character rather than an UDC character.
;
;    2) There is no provision to read data out of the Display. This saves
;       one I/O line, at the cost of having an image for the CWR (DIS_CWR)
;       and for the Flashing Status (FLASH_STA).
;
;
; Input parameters:
;   R0 = Address of character = FL/, A3, A2, A1, A0 (see above)
;   R7 = Character data (ASCII code, Flashing status or data to
;                         Control Word Register)
;
; Output:
;   P3.0 = DA0  (Display Address bit 0)
;   P3.1 = DA1  (Display Address bit 1)
;   P3.2 = DA2  (Display Address bit 2)
;   P3.3 = DA3  (Display Address bit 3)
;   P3.4 = DFL/ (Display Flashing Input)
;   P3.5 = DWR/ (Display Write Strobe)
;   P3.6 = DSC  (Display Serial Clock)
;   P3.7 = DSD  (Display Serial Data)
;
; Side effects: Acc is modified
;*****************************************************************************

Out_Char:
        MOV     OUT_BITS, #8        ;Number of bits to output
        MOV     A, R7               ;Move data to Acc
Out_Bit:    ; Output 8 bits of data to the shift register
        RRC     A                   ;Transfer bit to the Carry bit
        MOV     DSD, C              ; and then to Serial Data output
                ; Output Serial Clock Pulse
        SETB    DSC                 ;Set Display Serial Clock to high state
        CLR     DSC                 ;Set Display Serial Clock to low state
        DJNZ    OUT_BITS, Out_Bit   ;Loop until 8 bits are output

            ; ASCII code is now present on parallel outputs of shift register
            ;  so we can write it to the Display
                ; Output address (done now because we needed Acc for data)
        MOV     A, R0               ;Move address to Acc
        ORL     P3, A               ;Set 1's in bits 0-4 of P3
        ORL     A, #0E0h            ;Set bits 5-7 of Acc to 1 so that ORL
                                    ; has no effect on those bits
        ANL     P3, A               ;Clear 0's in bits 0-4 of P3
                ; Output Write Strobe
        CLR     DWR                 ;Set Write Strobe to low state
        SETB    DWR                 ;Set Write Strobe to high state
        RET



;*****************************************************************************
;
; Clock Subroutine
; ----------------
;
; Since flag GOT_TICK is set by the Timer 0 interrupt routine every 10 ms,
;  the Clock code is executed every 10 milliseconds.
;
; This subroutine is called once every go round the Background and also
;  from blind loops (currently the only blind loop is the 30 ms debouncing
;  delay, label WDbnc).
;
; Important warning: this mechanism works provided that the Background
;  ALWAYS executes in less time than the tick period (10 ms), except for
;  blind loops which have their own call to Clock, otherwise we could
;  miss a tick. In the current code, the maximum Background loop time is
;  about 2.5 ms, so we are safe.
;
; Clock fine tuning
; -----------------
;
; See the description at the beginning of the file.
;
; We must we skip one tick every CLOCK_ADJ seconds. We use a dedicated
;  variable SEC_ADJ to count the CLOCK_ADJ seconds since SEC_ADJ will
;  be cleared every time a tick is skipped, so that we can compare it
;  to CLOCK_ADJ again. However, due to the shortage in RAM space, we
;  do not have a dedicated hundredth of seconds counter. This does not
;  matter very much, because the error that this creates is constant
;  statistically, so it may be compensated for by adding 1 to the
;  calculated value of CLOCK_ADJ. It all boils down to the fact that
;  the tuning resolution is 1/CLOCK_ADJ. Since CLOCK_ADJ is larger than
;  100 in our case (it's unlikely to be much less in any case), and since
;  the frequency error itself is less than 100 ppm, the final accurracy
;  of the Clock can be made better than 1 ppm, i.e. less than 0.086 seconds
;  per day, which corresponds to about 30 seconds per year. Note: these
;  calculations do not take into account the frequency shift due to the
;  temperature... But we do our best for a crystal which is not in a
;  temperature controlled enclosure!
;
;
; Side effects: Acc, R0 and R7 are modified
;
;*****************************************************************************

Clock:
        JBC     GOT_TICK, Cft           ;Do Clock if 10 ms tick occurred and
        RET                             ; clear flag for next tick

                    ; Clock fine tuning
Cft:    MOV     A, #CLOCK_ADJ           ;If not counted CLOCK_ADJ second,
        CJNE    A, SEC_ADJ, Clk         ; normal processing
        MOV     SEC_ADJ, #0             ;Else clear CLOCK_ADJ for next time
        RET                             ; and skip this tick

Clk:                ; This is executed every 10 ms
        INC     HUNDTHS                 ;Increment hundredths of second
                                        ; for debouncing or for 400 ms delay
        INC     HUNDREDTHS              ;Increment hundredths of second
                                        ; of the Clock
        MOV     A, #100
        CJNE    A, HUNDREDTHS, XClk     ;If not counted one second, exit

Sec:    MOV     HUNDREDTHS, #0          ;Clear hundredths of second
        INC     SEC_ADJ                 ;Increment seconds for fine tuning
                                        ; (not in BCD format)
        MOV     R0, #SECONDS            ;Pass address of Seconds to Inc_BCD
        MOV     R7, #60h                ;Pass Modulo Seconds to Inc_BCD
        ACALL   Inc_BCD                 ;Increment seconds of the Clock
        JNZ     XClk                    ;If not counted one minute, exit

Min:    MOV     A, MINS                 ;Increment minutes for zone
        ADD     A, #1                   ; duration (not worth calling Inc_BCD
        DA      A                       ; because we increment modulo 100h,
        MOV     MINS, A                 ; i.e. 100 BCD)

        MOV     R0, #MINUTES            ;Pass address of Minutes to Inc_BCD
                ; Modulo Minutes (60) already in R7 from Increment Seconds
        ACALL   Inc_BCD                 ;Increment minutes of the Clock
        JNZ     XClk                    ;If not counted one hour, exit

Hour:   MOV     R0, #HOURS              ;Pass address of Minutes to Inc_BCD
        MOV     R7, #24h                ;Pass Modulo Hours to Inc_BCD
        ACALL   Inc_BCD                 ;Increment hours of the Clock
        JNZ     XClk                    ;If not counted one day, exit

Day:    INC     DAYS                    ;Increment days of the Clock
        MOV     A, #7
        CJNE    A, DAYS, XClk           ;If not counted 7 days, exit

Week:   MOV     DAYS, #0                ;Clear days

XClk:   RET



;*****************************************************************************
;
; Subroutine Nesting Tree
; -----------------------
;
; Drawing this tree permitted to calculate the maximum subroutine nesting
;  level which in turn permitted to calculate the stack size exactly, which
;  was essential, since all the RAM is used.
;
; The subroutines appear in the tree in the order in which they are called
;  in the source file (not the order in which they are called when the code
;  executes), but when a subroutine is called more than once at a given
;  level, only the first call is shown.
;
; The largest subroutine nesting level is 4 (see the last branch).
;
;
; Background --+-- Sch_All
;              |
;              +-- Clock --------- Inc_BCD
;              |
;              +-- Out_Char
;              |
;              +-- Hum_Msg
;              |
;              +-- Dis_Clk ----+-- Dis_Day ----+-- Day_Ch1
;              |               |               |
;              |               |               +-- Day_Ch2
;              |               |               |
;              |               |               +-- Out_Char
;              |               |
;              |               +-- BCD_ASC ------- Out_Char
;              |               |
;              |               +-- Out_Char
;              |
;              +-- Dis_Stt ----+-- Dis_Pr -------- Out_Char
;              |               |
;              |               +-- Out_Char
;              |               |
;              |               +-- BCD_ASC ------- Out_Char
;              |
;              +-- Set_Val ----+-- Inc_BCD
;              |               |
;              |               +-- Dec_BCD
;              |
;              +-- Dis_Dur ----+-- Dis_Pr -------- Out_Char
;              |               |
;              |               +-- Out_Char
;              |               |
;              |               +-- BCD_ASC ------- Out_Char
;              |
;              +-- Inc_BCD
;              |
;              +-- Dis_Sch ----+-- Dis_Day ----+-- Day_Ch1
;              |               |               |
;              |               |               +-- Day_Ch2
;              |               |               |
;              |               |               +-- Out_Char
;              |               |
;              |               +-- Out_Char
;              |               |
;              |               +-- Dis_Pr -------- Out_Char
;              |
;              +-- Dis_Pr -------- Out_Char
;              |
;              +-- Ctrl_Dur ------ Dis_Dur ----+-- Dis_Pr -------- Out_Char
;                                              |
;                                              +-- Out_Char
;                                              |
;                                              +-- BCD_ASC ------- Out_Char
;
;*****************************************************************************


END

